# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import copy
import shutil
import typing
import logging
import pathlib
import itertools

import rez.config
import rez.version
import rez.packages
import rez.package_maker
import rez.resolved_context

import rez_pip.pip
import rez_pip.utils
import rez_pip.plugins
import rez_pip.exceptions
from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger(__name__)


class NoPythonFound(rez_pip.exceptions.RezPipError):
    """
    Raised when no python package can be found.
    """


def createPackage(
    packageGroup: rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact],
    pythonVersion: rez.version.Version,
    installedWheelsDir: str,
    prefix: str | None = None,
    release: bool = False,
) -> None:
    _LOG.info(
        "Creating rez package for {0}".format(
            " + ".join(dist.name for dist in packageGroup.dists)
        )
    )

    rezNames = [
        rez_pip.utils.pythontDistributionNameToRez(dist.name)
        for dist in packageGroup.dists
    ]

    name = rezNames[0]
    version = rez_pip.utils.pythonDistributionVersionToRez(
        packageGroup.dists[0].version
    )

    requires = []
    variant_requires = []
    metadata: dict[str, typing.Any] = {}
    isPure = True
    for dist in packageGroup.dists:
        requirements = rez_pip.utils.getRezRequirements(dist, pythonVersion, [])
        if not metadata:
            # For now we only use the metadata from the first package. Far from ideal...
            metadata = requirements.metadata

        # TODO: Remove grouped packages (PySide-Addons, etc)
        requires += [
            require
            for require in requirements.requires
            if require not in requires
            # Check that the rez requirement isn't in the group name since it would be
            # an invalid requirement (because we merge them).
            and rez.version.Requirement(require).name not in rezNames[1:]
        ]
        variant_requires += [
            require
            for require in requirements.variant_requires
            if require not in variant_requires
            # Check that the rez requirement isn't in the group name since it would be
            # an invalid requirement (because we merge them).
            and rez.version.Requirement(require).name not in rezNames[1:]
        ]
        if isPure:
            isPure = metadata["is_pure_python"]

    if prefix:
        packagesPath = prefix
    else:
        packagesPath = (
            rez.config.config.release_packages_path
            if release
            else rez.config.config.local_packages_path
        )

    def make_root(variant: rez.packages.Variant, path: str) -> None:
        """Using distlib to iterate over all installed files of the current
        distribution to copy files to the target directory of the rez package
        variant
        """
        formattedRequirements = ", ".join(str(req) for req in variant.variant_requires)

        _LOG.info(
            rf"Installing {variant.qualified_package_name} \[{formattedRequirements}]"
        )
        for dist in packageGroup.dists:
            if not dist.files:
                raise RuntimeError(
                    f"{dist.name} package has no files registered! Something is wrong maybe?"
                )

            wheelsDirAbsolute = pathlib.Path(installedWheelsDir).resolve()
            for src in dist.files:
                srcAbsolute: pathlib.Path = typing.cast(
                    pathlib.Path, src.locate()
                ).resolve()
                dest = os.path.join(
                    path,
                    os.path.sep.join(
                        srcAbsolute.relative_to(wheelsDirAbsolute).parts[1:]
                    ),
                )
                # print(dest)
                if not os.path.exists(os.path.dirname(dest)):
                    os.makedirs(os.path.dirname(dest))

                _LOG.debug(f"Copying {str(srcAbsolute)!r} to {str(dest)!r}")
                shutil.copyfile(srcAbsolute, dest)
                shutil.copystat(srcAbsolute, dest)

    with rez.package_maker.make_package(
        name, packagesPath, make_root=make_root, skip_existing=True, warn_on_skip=False
    ) as pkg:
        pkg.version = version

        # requirements and variants
        if requires:
            pkg.requires = requires

        if variant_requires:
            pkg.variants = [variant_requires]

        # commands
        commands = ["env.PYTHONPATH.append('{root}/python')"]

        console_scripts = [
            ep.name for ep in dist.entry_points if ep.group == "console_scripts"
        ]
        if console_scripts:
            pkg.tools = console_scripts
            # TODO: Don't hardcode scripts here.
            commands.append("env.PATH.append('{root}/scripts')")

        pkg.commands = "\n".join(commands)

        # Make the package use hashed variants. This is required because we
        # can't control what ends up in its variants, and that can easily
        # include problematic chars (>, +, ! etc).
        # TODO: #672 (shortlinks for variants)
        pkg.hashed_variants = True

        pkg.pip = {
            "name": dist.name,
            "version": dist.version,
            "is_pure_python": isPure,
            "wheel_urls": packageGroup.downloadUrls,
            "rez_pip_version": importlib_metadata.version("rez-pip"),
        }

        # Take all the metadata that can be converted and put it
        # in the rez package definition.
        convertedMetadata, remainingMetadata = _convertMetadata(dist)
        for key, values in convertedMetadata.items():
            setattr(pkg, key, values)

        pkg.pip["metadata"] = remainingMetadata

        rez_pip.plugins.getHook().metadata(package=pkg)

    _LOG.info(
        f"[bold]Created {len(pkg.installed_variants)} variants and skipped {len(pkg.skipped_variants)}"
    )


def _convertMetadata(
    dist: importlib_metadata.Distribution,
) -> tuple[dict[str, typing.Any], dict[str, typing.Any]]:
    metadata: dict[str, typing.Any] = {}
    originalMetadata = copy.deepcopy(dist.metadata.json)
    del originalMetadata["metadata_version"]
    del originalMetadata["name"]
    del originalMetadata["version"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#summary
    if "Summary" in dist.metadata:
        metadata["summary"] = dist.metadata["Summary"]
        del originalMetadata["summary"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#description
    if "Description" in dist.metadata:
        metadata["description"] = dist.metadata["Description"]
        del originalMetadata["description"]

    authors = []

    # https://packaging.python.org/en/latest/specifications/core-metadata/#author
    if "Author" in dist.metadata:
        authors.append(dist.metadata["Author"])
        del originalMetadata["author"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#author-email
    if "Author-email" in dist.metadata:
        authors.extend(
            [email.strip() for email in dist.metadata["Author-email"].split(",")]
        )
        del originalMetadata["author_email"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#maintainer
    if "Maintainer" in dist.metadata:
        authors.append(dist.metadata["Maintainer"])
        del originalMetadata["maintainer"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#maintainer-email
    if "Maintainer-email" in dist.metadata:
        authors.extend(
            [email.strip() for email in dist.metadata["Maintainer-email"].split(",")]
        )
        del originalMetadata["maintainer_email"]

    if authors:
        metadata["authors"] = authors

    # https://packaging.python.org/en/latest/specifications/core-metadata/#license
    # Prefer the License field and fallback to classifiers if one is present.
    if "License" in dist.metadata:
        metadata["license"] = dist.metadata["License"]
        del originalMetadata["license"]
    elif "License-Expression" in dist.metadata:
        metadata["license"] = dist.metadata["License-Expression"]
        del originalMetadata["license_expression"]
    else:
        licenseClassifiers = [
            item.split("::")[-1].strip()
            for item in dist.metadata.get_all("classifier", [])
            if item.startswith("License ::")
        ]

        # I don't know what to do in this case, so just skip license if more than one
        # classifier is found.
        if len(licenseClassifiers) == 1:
            metadata["license"] = licenseClassifiers[0]

    helpLinks = []

    # https://packaging.python.org/en/latest/specifications/core-metadata/#home-page
    if "Home-page" in dist.metadata:
        helpLinks.append(["Home-page", dist.metadata["Home-page"]])
        del originalMetadata["home_page"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#project-url-multiple-use
    if "Project-URL" in dist.metadata:
        urls = [
            url.strip()
            for value in dist.metadata.get_all("Project-URL", failobj=[])
            for url in value.split(",")
        ]
        helpLinks.extend([list(entry) for entry in zip(urls[::2], urls[1::2])])
        del originalMetadata["project_url"]

    # https://packaging.python.org/en/latest/specifications/core-metadata/#download-url
    if "Download-URL" in dist.metadata:
        helpLinks.append(["Download-URL", dist.metadata["Download-URL"]])
        del originalMetadata["download_url"]

    if helpLinks:
        metadata["help"] = helpLinks

    return metadata, originalMetadata


def getPythonExecutables(
    range_: str | None, packageFamily: str = "python"
) -> dict[str, pathlib.Path]:
    """
    Get the available python executable from rez packages.

    :param range_: version specifier
    :param packageFamily: Name of the rez package family for the python package. This allows ot support PyPy, etc.
    :returns: Dict where the keys are the python versions and values are abolute paths to executables.
    """
    for family in rez.packages.iter_package_families():
        if family.name == "python":
            break
    else:
        raise NoPythonFound(f"No package family named {packageFamily!r} found")

    all_packages = sorted(
        rez.packages.iter_packages(
            packageFamily, range_=range_ if range_ != "latest" else None
        ),
        key=lambda x: x.version,
    )

    if not all_packages:
        raise NoPythonFound(
            f"No version found for package family {packageFamily!r} within requested range: {range_!r}"
        )

    packages: list[rez.packages.Package]
    if range_ == "latest":
        packages = [list(all_packages)[-1]]
    else:
        # Get the latest x.x (major+minor) and ignore anything else.
        # We don't want to return 3.7.8 AND 3.7.9 for example. It doesn't
        # make sense. We only need 3.7.x.
        groups = [
            list(group)
            for _, group in itertools.groupby(
                all_packages, key=lambda x: x.version.as_tuple()[:2]
            )
        ]
        # Note that "pkgs" is already in the right order since all_packages is sorted.
        packages = [pkgs[-1] for pkgs in groups]

    pythons: dict[str, pathlib.Path] = {}
    for package in packages:
        resolvedContext = rez.resolved_context.ResolvedContext(
            [f"{package.name}=={package.version}"]
        )

        # Make sure that system PATH doens't interfere with the "which" method.
        resolvedContext.append_sys_path = False

        for trimmedVersion in map(package.version.trim, [2, 1, 0]):
            path = resolvedContext.which(f"python{trimmedVersion}", parent_environ={})
            if path:
                pythons[str(package.version)] = pathlib.Path(path)
                break
        else:
            _LOG.warning(
                f"Failed to find a Python executable in the {package.qualified_name!r} rez package"
            )

    return pythons



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import rez_pip.plugins


def initializePluginManager(name: str):
    """Initialize a plugin manager and clear the cache before exiting the function

    :param name: Name of the plugin to load.
    """
    manager = rez_pip.plugins.getManager()
    for name, plugin in manager.list_name_plugin():
        if not name.startswith(f"rez-pip.plugins.{name}"):
            manager.unregister(plugin)

    try:
        yield manager
    finally:
        del manager
        rez_pip.plugins.getManager.cache_clear()



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import typing
import asyncio
import hashlib
import logging
import collections

import aiohttp
import rich.progress

import rez_pip.pip
import rez_pip.utils
from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger(__name__)
_lock = asyncio.Lock()


def downloadPackages(
    packageGroups: list[rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo]],
    dest: str,
) -> list[rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact]]:
    return asyncio.run(_downloadPackages(packageGroups, dest))


async def _downloadPackages(
    packageGroups: list[rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo]],
    dest: str,
) -> list[rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact]]:
    newPackageGroups: list[rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact]] = (
        []
    )
    someFailed = False

    async with aiohttp.ClientSession() as session:
        with rich.progress.Progress(
            "[progress.description]{task.description}",
            "[progress.percentage]{task.percentage:>3.0f}%",
            rich.progress.BarColumn(),
            rich.progress.DownloadColumn(),
            rich.progress.TransferSpeedColumn(),
            transient=True,
            console=rez_pip.utils.CONSOLE,
        ) as progress:
            tasks: dict[str, rich.progress.TaskID] = {}

            # Create all the download tasks first
            numPackages = 0
            for group in packageGroups:
                for package in group.packages:
                    if not package.isDownloadRequired():
                        continue

                    numPackages += 1
                    tasks[package.name] = progress.add_task(package.name)

            # Then create the "total" progress bar. This ensures that total is at the bottom.
            mainTask = progress.add_task(f"[bold]Total (0/{numPackages})", total=0)

            futures: list[
                typing.Coroutine[
                    typing.Any,
                    typing.Any,
                    rez_pip.pip.DownloadedArtifact | None,
                ]
            ] = []

            groupMapping: collections.defaultdict[int, list[str]] = (
                collections.defaultdict(list)
            )

            for index, group in enumerate(packageGroups):
                for package in group.packages:
                    wheelName: str = os.path.basename(package.download_info.url)
                    wheelPath = os.path.join(dest, wheelName)

                    groupMapping[index].append(package.name)

                    if not package.isDownloadRequired():

                        # Note the subtlety of having to pass variables in the function
                        # signature. We can't rely on the scoped variable.
                        async def _return_local(
                            _wheelPath: str, _package: rez_pip.pip.PackageInfo
                        ) -> rez_pip.pip.DownloadedArtifact:
                            return rez_pip.pip.DownloadedArtifact.from_dict(
                                {"_localPath": _wheelPath, **_package.to_dict()}
                            )

                        futures.append(_return_local(wheelPath, package))
                    else:
                        futures.append(
                            _download(
                                package,
                                session,
                                progress,
                                tasks[package.name],
                                mainTask,
                                wheelName,
                                wheelPath,
                            )
                        )

            artifacts = tuple(await asyncio.gather(*futures))

            if not all(artifacts):
                raise RuntimeError("Some wheels failed to be downloaded")

            artifacts = typing.cast(
                typing.Tuple[rez_pip.pip.DownloadedArtifact], artifacts
            )

            # Return artifacts in the same groups as they arrived in.
            # TODO: The amount of looping and gymnastic is a big code smell.
            for packageNames in groupMapping.values():
                newPackageGroups.append(
                    rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact](
                        tuple(
                            artifact
                            for artifact in artifacts
                            if artifact.name in packageNames
                        )
                    )
                )

    return newPackageGroups


def getSHA256(path: str) -> str:
    buf = bytearray(2**18)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)

    digestobj = hashlib.new("sha256")

    with open(path, "rb") as fd:
        while True:
            size = fd.readinto(buf)
            if size == 0:
                break  # EOF
            digestobj.update(view[:size])

    return digestobj.hexdigest()


async def _download(
    package: rez_pip.pip.PackageInfo,
    session: aiohttp.ClientSession,
    progress: rich.progress.Progress,
    taskID: rich.progress.TaskID,
    mainTaskID: rich.progress.TaskID,
    wheelName: str,
    wheelPath: str,
) -> rez_pip.pip.DownloadedArtifact | None:
    # TODO: Handle case where sha256 doesn't exist. We should also support the other supported
    # hash types.
    if (
        os.path.exists(wheelPath)
        and getSHA256(wheelPath) == package.download_info.archive_info.hashes["sha256"]
    ):
        _LOG.info(f"{wheelName} found in cache at {wheelPath!r}. Skipping download.")
    else:
        _LOG.debug(
            f"Downloading {package.name}-{package.version} from {package.download_info.url}"
        )

        async with session.get(
            package.download_info.url,
            headers={
                "Content-Type": "application/octet-stream",
                "User-Agent": f"rez-pip/{importlib_metadata.version('rez-pip')}",
            },
        ) as response:
            size = int(response.headers.get("content-length", 0))
            progress.update(taskID, total=size)

            async with _lock:
                mainTask = [task for task in progress.tasks if task.id == mainTaskID][0]

                progress.update(
                    mainTaskID,
                    total=typing.cast(int, mainTask.total) + size,
                )

            if response.status != 200:
                _LOG.error(
                    f"failed to download {package.download_info.url}: {response.status} - {response.reason}, {response.request_info}"
                )
                return None

            with open(wheelPath, "wb") as fd:
                async for chunk, asd in response.content.iter_chunks():
                    if not chunk:
                        break
                    fd.write(chunk)
                    progress.update(taskID, advance=len(chunk))
                    progress.update(mainTaskID, advance=len(chunk))

            _LOG.info(
                f"Downloaded {package.name}-{package.version} to {wheelPath!r} ({os.stat(wheelPath).st_size} bytes)"
            )

    progress.update(taskID, visible=False)

    total = len(progress.tasks) - 1
    async with _lock:
        completedItems = [task for task in progress.tasks if not task.visible]
        progress.update(
            mainTaskID, description=f"[bold]Total ({len(completedItems)}/{total})"
        )

    return rez_pip.pip.DownloadedArtifact.from_dict(
        {"_localPath": wheelPath, **package.to_dict()}
    )



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import time
import typing
import asyncio
import pathlib
import platform
import textwrap
import subprocess
import http.client

import pytest
import rattler
import rez.config
import rez.system
import rez.version
import rez.packages
import rez.package_bind
import rez.package_maker
import rez.package_remove

import rez_pip.utils

from . import utils

DATA_ROOT_DIR = os.path.join(os.path.dirname(__file__), "data")
CONDA_DIR = os.path.join(DATA_ROOT_DIR, "_conda")

phaseReportKey = pytest.StashKey[typing.Dict[str, pytest.CollectReport]]()


# Token from https://docs.pytest.org/en/latest/example/simple.html#making-test-result-information-available-in-fixtures
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item: pytest.Item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # store test results for each phase of a call, which can
    # be "setup", "call", "teardown"
    item.stash.setdefault(phaseReportKey, {})[rep.when] = rep


@pytest.fixture(scope="function", autouse=True)
def patchRichConsole(monkeypatch: pytest.MonkeyPatch):
    """Patch the rich console so that it doesn't wrap long lines"""
    monkeypatch.setattr(rez_pip.utils.CONSOLE, "width", 1000)


@pytest.fixture(scope="session")
def index(
    tmpdir_factory: pytest.TempdirFactory, printer_session: typing.Callable[[str], None]
) -> utils.PyPIIndex:
    """Build PyPI Index and return the path"""

    srcPackages = os.path.join(DATA_ROOT_DIR, "src_packages")

    indexPath = tmpdir_factory.mktemp("pypi_index").dirpath()

    for pkg in os.listdir(srcPackages):
        dest = indexPath.mkdir(pkg)
        printer_session(f"Building {pkg!r}...")
        wheel = utils.buildPackage(pkg, os.fspath(dest))
        printer_session(f"Built {pkg!r} at {wheel!r}")

    return utils.PyPIIndex(pathlib.Path(indexPath.strpath))


@pytest.fixture(scope="function")
def pypi(
    printer_session: typing.Callable[[str], None],
    index: utils.PyPIIndex,
    request: pytest.FixtureRequest,
) -> typing.Generator[str, None, None]:
    """Start a PyPI instance and return the URL to talk to it."""
    port = 45678
    host = "localhost"

    proc = subprocess.Popen(
        [
            "pypi-server",
            "run",
            os.fspath(index.path),
            f"--port={port}",
            f"--host={host}",
            "--disable-fallback",
            "--log-stream=stdout",
            "--hash-algo=sha256",  # Defaults to md5
            "-v",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    url = f"http://{host}:{port}"

    retries = 50
    while retries > 0:
        conn = http.client.HTTPConnection(f"{host}:{port}")
        try:
            conn.request("HEAD", "/health")
            response = conn.getresponse()
            if response is not None:
                print(response.read())
                yield url
                break
        except ConnectionRefusedError:
            time.sleep(0.1)
            retries -= 1

    proc.terminate()
    proc.wait()

    output = textwrap.indent(proc.stdout.read(), "\t\t")

    if proc.returncode and proc.returncode != -15 and platform.system() != "Windows":
        pytest.fail(
            f"pypi-server returned a non zero error code ({proc.returncode}):\n\n{proc.stdout.read()}"
        )

    if (
        request.node.stash[phaseReportKey].get("call")
        and request.node.stash[phaseReportKey]["call"].failed
    ):
        printer_session(f"pypi-server output:\n{output}")

    if not retries:
        raise RuntimeError("Failed to start pypi-server")


# Use function scope to make sure each test starts with a fresh default config
@pytest.fixture(scope="function", autouse=True)
def hardenRezConfig(tmp_path_factory: pytest.TempPathFactory):
    """Make sure that rez doesn't use any of the user configs"""

    # We can't just put None, so override with a dummy directory to make sure
    # that we don't pick up packages from anywhere else.
    tmpRepoPath = tmp_path_factory.mktemp("tmp_fake_rez_repo_do_not_use")

    # Do not load user configs
    defaultConfig = rez.config._create_locked_config(
        {
            "packages_path": [os.fspath(tmpRepoPath)],
            "release_packages_path": os.fspath(tmpRepoPath),
            "local_packages_path": os.fspath(tmpRepoPath),
        }
    )
    with rez.config._replace_config(defaultConfig):
        yield


@pytest.fixture(scope="function", autouse=True)
def resetRez():
    """Reset rez caches to make sure we don't leak anything between tests"""
    yield
    rez.system.system.clear_caches()


@pytest.fixture(scope="session")
def rezRepo() -> typing.Generator[str, None, None]:
    path = os.path.join(DATA_ROOT_DIR, "rez_repo")

    rez.package_bind.bind_package("platform", path=path, no_deps=True, quiet=True)
    rez.package_bind.bind_package("arch", path=path, no_deps=True, quiet=True)
    rez.package_bind.bind_package("os", path=path, no_deps=True, quiet=True)

    originalConfig = rez.config.config.copy()

    rez.config.config.packages_path = [path]
    rez.config.config.release_packages_path = path
    yield path

    rez.config.config.packages_path = originalConfig.packages_path
    rez.config.config.release_packages_path = originalConfig.release_packages_path


@pytest.fixture(
    scope="session",
    params=[
        pytest.param("3.8.20", marks=[pytest.mark.py38]),
        pytest.param("3.9.21", marks=[pytest.mark.py39]),
        pytest.param("3.11.11", marks=[pytest.mark.py311]),
    ],
)
def pythonRezPackage(
    request: pytest.FixtureRequest,
    rezRepo: str,
    printer_session: typing.Callable[[str], None],
) -> str:
    """
    Create a Python rez package and return Python version number.
    """
    version = typing.cast(str, request.param)

    def make_root(variant: rez.packages.Variant, path: str) -> None:
        """Using distlib to iterate over all installed files of the current
        distribution to copy files to the target directory of the rez package
        variant
        """
        printer_session(f"Creating rez package for Python {version} in {rezRepo!r}")

        dest = os.path.join(path, "python")

        printer_session(
            f"Installing Python {version} by creating a conda environment at {dest!r}"
        )

        asyncio.run(createCondaEnvironment(version, dest))

    try:
        with rez.package_maker.make_package(
            "python",
            rezRepo,
            make_root=make_root,
            skip_existing=True,
            warn_on_skip=False,
        ) as pkg:
            pkg.version = version

            commands = [
                "env.PATH.prepend('{root}/python/bin')",
            ]
            if platform.system() == "Windows":
                commands = [
                    "env.PATH.prepend('{root}/python/DLLs')",
                    "env.PATH.prepend('{root}/python/Library/bin')",
                    "env.PATH.prepend('{root}/python')",
                ]

            pkg.commands = "\n".join(commands)
    except Exception as exc:
        # If the creation fail, remove the package.
        # make_package doesn't do any cleanup if make_root fails...
        obj = rez.version.VersionedObject(f"python-{version}")
        rez.package_remove.remove_package(obj.name, obj.version, rezRepo)

        raise exc from None

    if pkg.skipped_variants:
        printer_session(
            f"Python {version} rez package already exists at {pkg.skipped_variants[0].uri}"
        )

    return version


async def createCondaEnvironment(pythonVersion: str, prefixPath: str):
    """Create a conda environment using py-rattler"""
    records = await rattler.solve(
        ["https://conda.anaconda.org/conda-forge"],
        [rattler.MatchSpec(f"python={pythonVersion}")],
        virtual_packages=rattler.VirtualPackage.detect(),
    )

    await rattler.install(
        records,
        prefixPath,
        cache_dir=os.path.join(CONDA_DIR, "pkgs"),
    )



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import math
import typing
import logging
import contextlib
import collections.abc
import logging.handlers

import patch_ng

import rez_pip.utils
import rez_pip.compat
import rez_pip.plugins
import rez_pip.exceptions
import rez_pip.data.patches
from rez_pip.compat import importlib_metadata


_LOG = logging.getLogger(__name__)


class PatchError(rez_pip.exceptions.RezPipError):
    pass


def getBuiltinPatchesDir() -> str:
    """Get the built-in patches directory"""
    return os.path.dirname(rez_pip.data.patches.__file__)


@contextlib.contextmanager
def logIfErrorOrRaises() -> typing.Generator[None, None, None]:
    """
    Log patch_ng logs if any error is logged or if the wrapped body raises.
    Very slightly inspired by https://docs.python.org/3/howto/logging-cookbook.html#buffering-logging-messages-and-outputting-them-conditionally

    We basically don't want any logs from patch_ng is everything worked. We only
    want logs when something wrong happens.
    """
    patch_ng.debugmode = True
    logger = logging.getLogger("patch_ng")
    initialLevel = logger.level
    logger.setLevel(logging.DEBUG)

    handler = logging.handlers.MemoryHandler(
        math.inf,  # type: ignore[arg-type]
        flushLevel=logging.ERROR,
        target=logging.getLogger("rez_pip").handlers[0],
    )
    handler.setFormatter(logging.Formatter("%(name)s %(levelname)8s %(message)s"))

    logger.addHandler(handler)

    try:
        yield
    except Exception as exc:
        handler.flush()
        raise exc from None
    finally:
        patch_ng.debugmode = False
        logger.setLevel(initialLevel)
        logger.removeHandler(handler)


def patch(dist: importlib_metadata.Distribution, path: str) -> None:
    """Patch an installed package (wheel)"""
    _LOG.debug(f"[bold]Attempting to patch {dist.name!r} at {path!r}")
    patchesGroups: collections.abc.Sequence[collections.abc.Sequence[str]] = (
        rez_pip.plugins.getHook().patches(dist=dist, path=path)
    )

    # Flatten the list
    patches = [path for group in patchesGroups for path in group]

    if not patches:
        _LOG.debug(f"No patches found")
        return

    _LOG.info(f"Applying {len(patches)} patches for {dist.name!r} at {path!r}")

    for patch in patches:
        _LOG.info(f"Applying patch {patch!r} on {path!r}")

        if not os.path.isabs(patch):
            raise PatchError(f"{patch!r} is not an absolute path")

        if not os.path.exists(patch):
            raise PatchError(f"Patch at {patch!r} does not exist")

        patchset = patch_ng.fromfile(patch)
        with logIfErrorOrRaises():
            if not patchset.apply(root=path):
                # A logger that only gets flushed on demand would be better...
                raise PatchError(f"Failed to apply patch {patch!r} on {path!r}")



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

import sys


def run():
    # We want to test that the executable is really the correct one, the one we expect.
    # So printing it will allow us to compare it.
    sys.stdout.write(sys.executable)



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""Needed to corectly track coverage in subprocesses from the integration tests"""

import coverage

coverage.process_startup()



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import typing
import hashlib
import pathlib
import unittest.mock

import pytest
import aiohttp

import rez_pip.pip
import rez_pip.download
from rez_pip.compat import importlib_metadata


@pytest.fixture(scope="module", autouse=True)
def rezPipVersion():
    with unittest.mock.patch.object(
        importlib_metadata, "version", return_value="1.2.3.4.5"
    ):
        yield


class Package:
    def __init__(self, name: str, content: str, local: bool):
        self.name = name
        self.content = content
        self.local = local


class Group:
    def __init__(self, packages: typing.List[Package]):
        self.packages = packages

    def getPackage(self, name: str) -> Package:
        for package in self.packages:
            if package.name == name:
                return package
        raise KeyError(name)


@pytest.mark.parametrize(
    "groups",
    [
        [Group([Package("package-a", "package-a data", False)])],
        [
            Group([Package("package-a", "package-a data", False)]),
            Group([Package("package-b", "package-b data", False)]),
        ],
    ],
    ids=["one-group-with-one-package", "multiple-groups-with-one-package"],
)
def test_download(groups: typing.List[Group], tmp_path: pathlib.Path):
    sideEffects = tuple()
    for group in groups:
        for package in group.packages:
            mockedContent = unittest.mock.MagicMock()
            mockedContent.return_value.__aiter__.return_value = [
                [
                    package.content.encode("utf-8"),
                    None,
                ]
            ]

            sideEffects += (
                unittest.mock.Mock(
                    headers={"content-length": 100},
                    status=200,
                    content=unittest.mock.Mock(iter_chunks=mockedContent),
                ),
            )

    mockedGet = unittest.mock.AsyncMock()
    mockedGet.__aenter__.side_effect = sideEffects

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet

        _groups = []
        for group in groups:
            infos = []
            for package in group.packages:
                infos.append(
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(
                            name=package.name, version="1.0.0"
                        ),
                        download_info=rez_pip.pip.DownloadInfo(
                            url=f"https://example.com/{package.name}.whl",
                            archive_info=rez_pip.pip.ArchiveInfo("hash", {}),
                        ),
                        is_direct=True,
                        requested=True,
                    )
                )
            _groups.append(rez_pip.pip.PackageGroup(infos))

        new_groups = rez_pip.download.downloadPackages(_groups, os.fspath(tmp_path))

    assert len(new_groups) == len(groups)
    assert sum(len(group.packages) for group in new_groups) == sum(
        len(group.packages) for group in groups
    )

    wheelsMapping = {
        package.name: package.path for group in new_groups for package in group.packages
    }

    for group in groups:
        for package in group.packages:
            with open(wheelsMapping[package.name]) as fd:
                content = fd.read()
            assert content == package.content

    assert mocked.call_args_list == [
        unittest.mock.call(
            f"https://example.com/{package.name}.whl",
            headers={
                "Content-Type": "application/octet-stream",
                "User-Agent": "rez-pip/1.2.3.4.5",
            },
        )
        for group in groups
        for package in group.packages
    ]


def test_download_skip_local(tmp_path: pathlib.Path):
    """Test that wheels are not downloaded if they are local wheels"""
    groups = [
        rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo](
            [
                rez_pip.pip.PackageInfo(
                    metadata=rez_pip.pip.Metadata(name="package-a", version="1.0.0"),
                    download_info=rez_pip.pip.DownloadInfo(
                        url="file:///example.com/package-a",
                        archive_info=rez_pip.pip.ArchiveInfo("hash-a", {}),
                    ),
                    is_direct=True,
                    requested=True,
                )
            ]
        )
    ]

    mockedGet = unittest.mock.AsyncMock()

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet
        wheels = rez_pip.download.downloadPackages(groups, os.fspath(tmp_path))

    assert not mocked.called
    data = rez_pip.pip.PackageGroup(
        (
            rez_pip.pip.DownloadedArtifact.from_dict(
                {
                    **groups[0].packages[0].to_dict(),
                    "_localPath": os.path.join(os.fspath(tmp_path), "package-a"),
                }
            ),
        )
    )
    assert wheels == [data]


def test_download_multiple_packages_with_failure(tmp_path: pathlib.Path):
    """
    Test that a failure in one package does not prevent other
    packages from being downloaded
    """
    mockedContent = unittest.mock.MagicMock()
    mockedContent.return_value.__aiter__.return_value = [
        [
            b"package-a data",
            None,
        ]
    ]

    mockedGet = unittest.mock.AsyncMock()
    mockedGet.__aenter__.side_effect = (
        unittest.mock.Mock(
            headers={"content-length": 100},
            status=200,
            content=unittest.mock.Mock(iter_chunks=mockedContent),
        ),
        unittest.mock.Mock(
            headers={"content-length": 100},
            status=400,
            reason="Expected to fail",
            request_info={"key": "here"},
            content=unittest.mock.Mock(iter_chunks=mockedContent),
        ),
    )

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet
        with pytest.raises(RuntimeError):
            rez_pip.download.downloadPackages(
                [
                    rez_pip.pip.PackageGroup(
                        [
                            rez_pip.pip.PackageInfo(
                                metadata=rez_pip.pip.Metadata(
                                    name="package-a", version="1.0.0"
                                ),
                                download_info=rez_pip.pip.DownloadInfo(
                                    url="https://example.com/package-a",
                                    archive_info=rez_pip.pip.ArchiveInfo("hash-a", {}),
                                ),
                                is_direct=True,
                                requested=True,
                            )
                        ]
                    ),
                    rez_pip.pip.PackageGroup(
                        [
                            rez_pip.pip.PackageInfo(
                                metadata=rez_pip.pip.Metadata(
                                    name="package-b", version="1.0.0"
                                ),
                                download_info=rez_pip.pip.DownloadInfo(
                                    url="https://example.com/package-b",
                                    archive_info=rez_pip.pip.ArchiveInfo("hash-b", {}),
                                ),
                                is_direct=True,
                                requested=True,
                            )
                        ]
                    ),
                ],
                os.fspath(tmp_path),
            )

        # Check that package-a was downloaded even if even if package-b failed.
        with open(tmp_path / "package-a") as fd:
            assert fd.read() == "package-a data"

        assert mocked.call_args_list == [
            unittest.mock.call(
                "https://example.com/package-a",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            unittest.mock.call(
                "https://example.com/package-b",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]


def test_download_reuse_if_same_hash(tmp_path: pathlib.Path):
    """Test that wheels are re-used if the sha256 matches"""
    sideEffects = tuple()
    groups = []

    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        groups.append(
            rez_pip.pip.PackageGroup(
                [
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                        download_info=rez_pip.pip.DownloadInfo(
                            url=f"https://example.com/{package}.whl",
                            archive_info=rez_pip.pip.ArchiveInfo(
                                "hash-a", {"sha256": hash.hexdigest()}
                            ),
                        ),
                        is_direct=True,
                        requested=True,
                    )
                ]
            )
        )

        mockedContent = unittest.mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            unittest.mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=unittest.mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet1 = unittest.mock.AsyncMock()
    mockedGet1.__aenter__.side_effect = sideEffects

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet1

        rez_pip.download.downloadPackages(groups, str(tmp_path))

        assert mocked.call_args_list == [
            unittest.mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            unittest.mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    groups = []
    # package-b will be re-used
    for package in ["package-c", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        groups.append(
            rez_pip.pip.PackageGroup(
                [
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                        download_info=rez_pip.pip.DownloadInfo(
                            url=f"https://example.com/{package}.whl",
                            archive_info=rez_pip.pip.ArchiveInfo(
                                "hash-a", {"sha256": hash.hexdigest()}
                            ),
                        ),
                        is_direct=True,
                        requested=True,
                    )
                ]
            )
        )

        mockedContent = unittest.mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            unittest.mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=unittest.mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet2 = unittest.mock.AsyncMock()
    mockedGet2.__aenter__.side_effect = sideEffects

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet2

        wheels = rez_pip.download.downloadPackages(groups, str(tmp_path))

        assert mocked.call_args_list == [
            unittest.mock.call(
                "https://example.com/package-c.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    assert len(wheels) == 2
    assert len(wheels[0].packages) == 1
    assert len(wheels[1].packages) == 1


def test_download_redownload_if_hash_changes(tmp_path: pathlib.Path):
    """Test that wheels are re-downloaded if the sha256 changes"""
    sideEffects = tuple()
    groups = []

    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        hash = hashlib.new("sha256")
        hash.update(content)

        groups.append(
            rez_pip.pip.PackageGroup(
                [
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                        download_info=rez_pip.pip.DownloadInfo(
                            url=f"https://example.com/{package}.whl",
                            archive_info=rez_pip.pip.ArchiveInfo(
                                "hash-a", {"sha256": hash.hexdigest()}
                            ),
                        ),
                        is_direct=True,
                        requested=True,
                    )
                ]
            )
        )

        mockedContent = unittest.mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            unittest.mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=unittest.mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet1 = unittest.mock.AsyncMock()
    mockedGet1.__aenter__.side_effect = sideEffects

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet1

        rez_pip.download.downloadPackages(groups, str(tmp_path))

        assert mocked.call_args_list == [
            unittest.mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            unittest.mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    groups = []
    for package in ["package-a", "package-b"]:
        content = f"{package} data".encode("utf-8")

        groups.append(
            rez_pip.pip.PackageGroup(
                [
                    rez_pip.pip.PackageInfo(
                        metadata=rez_pip.pip.Metadata(name=package, version="1.0.0"),
                        download_info=rez_pip.pip.DownloadInfo(
                            url=f"https://example.com/{package}.whl",
                            archive_info=rez_pip.pip.ArchiveInfo(
                                #
                                # Bad sha256. This will trigger a new download
                                #
                                "hash-a",
                                {"sha256": "asd"},
                            ),
                        ),
                        is_direct=True,
                        requested=True,
                    )
                ]
            )
        )

        mockedContent = unittest.mock.MagicMock()
        mockedContent.return_value.__aiter__.return_value = [
            [
                content,
                None,
            ]
        ]

        sideEffects += (
            unittest.mock.Mock(
                headers={"content-length": 100},
                status=200,
                content=unittest.mock.Mock(iter_chunks=mockedContent),
            ),
        )

    mockedGet2 = unittest.mock.AsyncMock()
    mockedGet2.__aenter__.side_effect = sideEffects

    with unittest.mock.patch.object(aiohttp.ClientSession, "get") as mocked:
        mocked.return_value = mockedGet2

        wheels = rez_pip.download.downloadPackages(groups, str(tmp_path))

        assert mocked.call_args_list == [
            unittest.mock.call(
                "https://example.com/package-a.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
            unittest.mock.call(
                "https://example.com/package-b.whl",
                headers={
                    "Content-Type": "application/octet-stream",
                    "User-Agent": "rez-pip/1.2.3.4.5",
                },
            ),
        ]

    assert len(wheels) == 2
    assert len(wheels[0].packages) == 1
    assert len(wheels[1].packages) == 1



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pluggy

import rez_pip.plugins


def test_getManager():
    assert isinstance(rez_pip.plugins.getManager(), pluggy.PluginManager)


def test_getHook():
    assert isinstance(rez_pip.plugins.getHook(), pluggy.HookRelay)


def test_getHookImplementations():
    implementations = rez_pip.plugins._getHookImplementations()
    assert implementations == {
        "rez_pip.PySide6": [
            "cleanup",
            "groupPackages",
            "patches",
            "postPipResolve",
            "prePipResolve",
        ],
        "rez_pip.shiboken6": ["cleanup"],
    }



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import pathlib
import unittest.mock

import pytest

import rez_pip.pip
import rez_pip.plugins
import rez_pip.exceptions

from . import utils


@pytest.fixture(scope="module", autouse=True)
def setupPluginManager():
    yield utils.initializePluginManager("pyside6")


@pytest.mark.parametrize(
    "packages",
    [
        ("asd",),
        ("pyside6",),
        ("PysiDe6",),
        ("pyside6", "pyside6-addons"),
        ("pyside6", "pyside6-essentials"),
        ("pyside6", "pyside6-essentials", "pyside6-addons"),
        ("pyside6", "pyside6-addons", "asdasdad"),
    ],
)
def test_prePipResolve_noop(packages: tuple[str, ...]):
    rez_pip.plugins.getHook().prePipResolve(packages=packages)


@pytest.mark.parametrize("packages", [("pyside6-addons",), ("PysiDe6_essentials",)])
def test_prePipResolve_raises(packages: tuple[str, ...]):
    with pytest.raises(rez_pip.exceptions.RezPipError):
        rez_pip.plugins.getHook().prePipResolve(packages=packages)


def fakePackage(name: str, **kwargs) -> unittest.mock.Mock:
    value = unittest.mock.MagicMock()
    value.configure_mock(name=name, **kwargs)
    return value


@pytest.mark.parametrize(
    "packages",
    [
        (fakePackage("asd"),),
        (fakePackage("pyside6"),),
        (fakePackage("PysiDe6"),),
        (fakePackage("pyside6"), fakePackage("pyside6-addons")),
        (fakePackage("pyside6"), fakePackage("pyside6-essentials")),
        (
            fakePackage("pyside6"),
            fakePackage("pyside6-essentials"),
            fakePackage("pyside6-addons"),
        ),
        (
            fakePackage("pyside6"),
            fakePackage("pyside6-addons"),
            fakePackage("asdasdad"),
        ),
    ],
)
def test_postPipResolve_noop(packages: tuple[str, ...]):
    rez_pip.plugins.getHook().postPipResolve(packages=packages)


@pytest.mark.parametrize(
    "packages",
    [
        (fakePackage("pyside6-addons"),),
        (fakePackage("PysiDe6_essentials"),),
        (fakePackage("PysiDe6_essentials"), fakePackage("asd")),
    ],
)
def test_postPipResolve_raises(packages: tuple[str, ...]):
    with pytest.raises(rez_pip.exceptions.RezPipError):
        rez_pip.plugins.getHook().postPipResolve(packages=packages)


@pytest.mark.parametrize(
    "packages",
    [[fakePackage("asd")]],
)
def test_groupPackages_noop(packages: list[str]):
    assert rez_pip.plugins.getHook().groupPackages(packages=packages) == [
        [rez_pip.pip.PackageGroup(tuple())]
    ]


class FakePackageInfo:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version

    def __eq__(self, value):
        return self.name == value.name and self.version == value.version


@pytest.mark.parametrize(
    "packages,expectedGroups",
    [
        [
            [fakePackage("pyside6", version="1")],
            [[rez_pip.pip.PackageGroup((FakePackageInfo("pyside6", "1"),))]],
        ],
        [
            [
                fakePackage("pyside6", version="1"),
                fakePackage("pyside6_addons", version="1"),
            ],
            [
                [
                    rez_pip.pip.PackageGroup(
                        (
                            FakePackageInfo("pyside6", "1"),
                            FakePackageInfo("pyside6_addons", "1"),
                        )
                    )
                ]
            ],
        ],
        [
            [
                fakePackage("pyside6", version="1"),
                fakePackage("pyside6_essentials", version="1"),
            ],
            [
                [
                    rez_pip.pip.PackageGroup(
                        (
                            FakePackageInfo("pyside6", "1"),
                            FakePackageInfo("pyside6_essentials", "1"),
                        )
                    )
                ]
            ],
        ],
        [
            [
                fakePackage("pyside6", version="1"),
                fakePackage("pyside6_essentials", version="1"),
                fakePackage("pyside6-Addons", version="1"),
            ],
            [
                [
                    rez_pip.pip.PackageGroup(
                        (
                            FakePackageInfo("pyside6", "1"),
                            FakePackageInfo("pyside6_essentials", "1"),
                            FakePackageInfo("pyside6-Addons", "1"),
                        )
                    )
                ]
            ],
        ],
        [
            [
                fakePackage("pyside6", version="1"),
                fakePackage("asdasd", version=2),
                fakePackage("pyside6_essentials", version="1"),
                fakePackage("pyside6-Addons", version="1"),
            ],
            [
                [
                    rez_pip.pip.PackageGroup(
                        (
                            FakePackageInfo("pyside6", "1"),
                            FakePackageInfo("pyside6_essentials", "1"),
                            FakePackageInfo("pyside6-Addons", "1"),
                        )
                    )
                ]
            ],
        ],
    ],
)
def test_groupPackages(
    packages: list[str], expectedGroups: list[rez_pip.pip.PackageGroup]
):
    data = rez_pip.plugins.getHook().groupPackages(packages=packages)
    assert data == expectedGroups


@pytest.mark.parametrize("package", [fakePackage("asd")])
def test_cleanup_noop(package, tmp_path: pathlib.Path):
    (tmp_path / "python" / "shiboken6").mkdir(parents=True)
    (tmp_path / "python" / "shiboken6_generator").mkdir(parents=True)

    rez_pip.plugins.getHook().cleanup(dist=package, path=tmp_path)

    assert (tmp_path / "python" / "shiboken6").exists()
    assert (tmp_path / "python" / "shiboken6_generator").exists()


@pytest.mark.parametrize(
    "package,expectedPaths",
    [
        [fakePackage("pyside6"), ["shiboken6", "shiboken6_generator"]],
        [
            fakePackage("pyside6_essentials"),
            [
                "shiboken6",
                "shiboken6_generator",
                os.path.join("PySide6", "__init__.py"),
            ],
        ],
        [
            fakePackage("PySiDe6-AddoNs"),
            [
                "shiboken6",
                "shiboken6_generator",
                os.path.join("PySide6", "__init__.py"),
            ],
        ],
    ],
)
def test_cleanup(package, expectedPaths: list[str], tmp_path: pathlib.Path):
    actions = rez_pip.plugins.getHook().cleanup(dist=package, path=tmp_path)

    expectedActions = []
    for path in expectedPaths:
        expectedActions.append(
            rez_pip.plugins.CleanupAction(
                "remove",
                str(tmp_path / "python" / path),
            )
        )
    assert actions == [[], expectedActions]



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""PyPI/python package ingester/converter for the rez package manager"""

from __future__ import annotations

import argparse


command_behavior = {
    "hidden": False,  # optional: bool
    "arg_mode": "grouped",  # optional: None, "passthrough", "grouped"
}


def setup_parser(parser: argparse.ArgumentParser) -> None:
    import rez_pip.cli

    rez_pip.cli._setupParser(parser, fromRez=True)


def command(
    opts: argparse.Namespace,
    _: argparse.ArgumentParser,
    extra_arg_groups: list[list[str]],
) -> int:
    import rez_pip.cli

    pipArgs = [arg for group in extra_arg_groups for arg in group]
    return rez_pip.cli.run(args=opts, pipArgs=pipArgs)


def register_plugin():  # type: ignore
    # Defined here to avoid cirlular imports
    import rez.command

    class RezPip(rez.command.Command):  # type: ignore
        """asd"""

    return RezPip



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import re
import sys
import pathlib
import platform
import subprocess

import pytest
import rez.version
import rez.packages
import rez.resolved_context


@pytest.mark.skipif(
    platform.system() == "Darwin" and not os.environ.get("CI"),
    reason="Skipping when running locally on macOS",
)
@pytest.mark.integration
def test_python_packages(pythonRezPackage: str, rezRepo: str):
    """Test that the python rez packages created by the pythonRezPackage fixture are functional"""

    package = rez.packages.get_package("python", pythonRezPackage, paths=[rezRepo])
    assert package

    ctx = rez.resolved_context.ResolvedContext(
        [package.qualified_name], package_paths=[rezRepo]
    )

    executableName = f"python{str(package.version).split('.')[0]}"

    expectedExecutablePath = os.path.join(
        rezRepo,
        "python",
        str(package.version),
        "python",
    )

    if platform.system() == "Windows":
        executableName = "python.exe"
        expectedExecutablePath = os.path.join(expectedExecutablePath, executableName)
    else:
        expectedExecutablePath = os.path.join(
            expectedExecutablePath, "bin", executableName
        )

    code, stdout, _ = ctx.execute_shell(
        command=[executableName, "-c", "import sys; print(sys.executable)"],
        block=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    assert code == 0
    assert stdout.strip().lower() == expectedExecutablePath.lower()

    code, stdout, _ = ctx.execute_shell(
        command=[executableName, "--version"],
        block=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    assert code == 0
    assert re.findall(r"\d\.\d+\.\d+", stdout, flags=re.MULTILINE)[0] == str(
        package.version
    )


@pytest.mark.parametrize(
    "packagesToInstall,imports", [[["PySide6"], ["PySide6"]]], ids=["PySide6"]
)
def test_installs(
    pythonRezPackage: str,
    rezRepo: str,
    packagesToInstall: list[str],
    imports: list[str],
    tmp_path: pathlib.Path,
    capsys: pytest.CaptureFixture,
):
    """End to end integration test"""
    command = [
        sys.executable,
        "-m",
        "rez_pip",
        *packagesToInstall,
        "--prefix",
        os.fspath(tmp_path),
        "--python-version",
        pythonRezPackage,
    ]

    env = {
        "REZ_PACKAGES_PATH": os.pathsep.join([rezRepo, os.fspath(tmp_path)]),
        "REZ_DISABLE_HOME_CONFIG": "1",
        "COVERAGE_PROCESS_START": os.path.join(
            os.path.dirname(os.path.dirname(__file__)), ".coveragerc"
        ),
        "PYTHONPATH": os.path.dirname(__file__),
    }
    if platform.system() == "Windows":
        # https://stackoverflow.com/a/64706392
        env["SYSTEMROOT"] = os.environ["SYSTEMROOT"]
        # Needed for getpass.getuser to work on Windows.
        env["USERNAME"] = os.environ["USERNAME"]

    with capsys.disabled():
        subprocess.check_call(command, env=env)

    ctx = rez.resolved_context.ResolvedContext(
        packagesToInstall + [f"python-{pythonRezPackage}"],
        package_paths=[rezRepo, os.fspath(tmp_path)],
    )
    assert ctx.status == rez.resolved_context.ResolverStatus.solved

    code, stdout, stderr = ctx.execute_shell(
        command=[
            "python",
            "-c",
            f"import {','.join(imports)}; [print(i.__path__[0]) for i in [{','.join(imports)}]]",
        ],
        block=True,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print(stdout)
    print("****")
    print(stderr)
    assert code == 0

    for path in stdout.strip().split("\n"):
        assert path.lower().startswith(
            os.fspath(tmp_path).lower()
        ), f"{path!r} does not start with {os.fspath(tmp_path)!r}"

    assert not stderr



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import sys
import json
import shutil
import logging
import argparse
import textwrap
import tempfile
import itertools
import subprocess

import rich.text
import rich.panel
import rich.table
import rez.version
import rich.markup
import rich.logging

import rez_pip.pip
import rez_pip.rez
import rez_pip.data
import rez_pip.patch
import rez_pip.utils
import rez_pip.plugins
import rez_pip.install
import rez_pip.download
import rez_pip.exceptions
from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger("rez_pip.cli")

__all__ = ["run"]


def __dir__() -> list[str]:
    return __all__


def _setupParser(
    parser: argparse.ArgumentParser, fromRez: bool = False
) -> argparse.ArgumentParser:
    parser.add_argument("packages", nargs="*", help="Packages to install.")

    generalGroup = parser.add_argument_group(title="general options")
    generalGroup.add_argument(
        "-r",
        "--requirement",
        action="append",
        metavar="<file>",
        help="Install from the given requirements file. This option can be used multiple times.",
    )
    generalGroup.add_argument(
        "-c",
        "--constraint",
        action="append",
        metavar="<file>",
        help="Constrain versions using the given constraints file. This option can be used multiple times.",
    )
    generalGroup.add_argument(
        "-p",
        "--prefix",
        metavar="<path>",
        help="Custom repository path (can be any directory, even non rez repository path) (default: configured local_packages_path)",
    )
    generalGroup.add_argument(
        "--release",
        action="store_true",
        help="Release the converted packages (default: configured release_packages_path)",
    )

    generalGroup.add_argument(
        "--python-version",
        # 3.7+ because that's what pip supports
        default="3.7+",
        metavar="<version>",
        help="Range of python versions. It can also be a single version, any valid rez version specifier or range or 'latest' (default: 3.7+)",
    )
    generalGroup.add_argument(
        "--pip",
        default=rez_pip.pip.getBundledPip(),
        metavar="<path>",
        help="Standalone pip (https://pip.pypa.io/en/stable/installation/#standalone-zip-application) (default: bundled).",
    )

    # Only needed to tests
    generalGroup.add_argument("--noop", action="store_true", help=argparse.SUPPRESS)

    # Rez injects --version and --help... So we have to
    # avoid setting them when run from rez.
    if not fromRez:
        # Manually define just to keep the style consistent (capital letters, dot, etc.)
        generalGroup.add_argument(
            "-h", "--help", action="help", help="Show this help message and exit."
        )

    generalGroup.add_argument(
        "--plugin-version",
        action="version",
        version=importlib_metadata.version(__package__),
    )

    debugGroup = parser.add_argument_group(title="debug options")
    debugGroup.add_argument(
        "-l",
        "--log-level",
        default="info",
        choices=["info", "debug", "warning", "error"],
        help="Logging level.",
    )

    debugGroup.add_argument(
        "--keep-tmp-dirs",
        action="store_true",
        help="Keep some temporary directories at the end of the process for further inspection.",
    )

    debugGroup.add_argument(
        "--debug-info",
        action="store_true",
        help="Print debug information that you can use when reporting an issue on GitHub.",
    )

    debugGroup.add_argument(
        "--list-plugins", action="store_true", help="List all registered plugins"
    )

    parser.usage = f"""

  %(prog)s [options] <package(s)>
  %(prog)s <package(s)> [-- [pip options]]
"""
    return parser


def _createParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Ingest and convert python packages to rez packages.",
        prog=__package__.replace("_", "-"),
        add_help=False,
    )
    return _setupParser(parser)


def _parseArgs(
    args: list[str],
) -> tuple[argparse.Namespace, list[str]]:
    parser = _createParser()

    knownArgs = []
    pipArgs = []
    if "--" in args:
        # anything after -- will be passed as is to pip.
        splitIndex = args.index("--")
        knownArgs = args[:splitIndex]
        pipArgs = args[splitIndex + 1 :]
    else:
        knownArgs = args

    ownArgs = parser.parse_args(knownArgs)

    return ownArgs, pipArgs


def _validateArgs(args: argparse.Namespace) -> None:
    if not args.pip.endswith(".pyz"):
        raise rez_pip.exceptions.RezPipError(
            f"[bold red]{args.pip!r} does not look like a valid zipapp. A zipapp should end with '.pyz'.[/]\n\n"
            "  [grey74]Standalone pip documentation[/]: https://pip.pypa.io/en/stable/installation/#standalone-zip-application\n"
            "  [grey74]zipapp documentation[/]: https://docs.python.org/3/library/zipapp.html"
        )

    if not os.path.exists(args.pip):
        raise rez_pip.exceptions.RezPipError(rf"zipapp at {args.pip!r} does not exist")

    if not args.packages and not args.requirement:
        raise rez_pip.exceptions.RezPipError(
            "no packages were passed and --requirements was not used. At least one of them must be passed."
        )


def _run(args: argparse.Namespace, pipArgs: list[str], pipWorkArea: str) -> None:
    pythonVersions = rez_pip.rez.getPythonExecutables(
        args.python_version, packageFamily="python"
    )

    if not pythonVersions:
        raise rez_pip.exceptions.RezPipError(
            f'No "python" package found within the range {args.python_version!r}.'
        )

    for pythonVersion, pythonExecutable in pythonVersions.items():
        _LOG.info(
            f"[bold underline]Installing requested packages for Python {pythonVersion}"
        )

        wheelsDir = os.path.join(pipWorkArea, "wheels")
        os.makedirs(wheelsDir, exist_ok=True)

        # Suffix with the python version because we loop over multiple versions,
        # and package versions, content, etc can differ for each Python version.
        installedWheelsDir = os.path.join(pipWorkArea, "installed", pythonVersion)
        os.makedirs(installedWheelsDir, exist_ok=True)

        with rez_pip.utils.CONSOLE.status(
            f"[bold]Resolving dependencies for {rich.markup.escape(', '.join(args.packages))} (python-{pythonVersion})"
        ):
            packages = rez_pip.pip.getPackages(
                args.packages,
                args.pip,
                pythonVersion,
                os.fspath(pythonExecutable),
                args.requirement or [],
                args.constraint or [],
                pipArgs,
            )

        _LOG.info(f"Resolved {len(packages)} dependencies for python {pythonVersion}")
        _packageGroups: list[rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo]] = list(
            itertools.chain(*rez_pip.plugins.getHook().groupPackages(packages=packages))  # type: ignore[arg-type]
        )

        # TODO: Verify that no packages are in two or more groups? It should theorically
        # not be possible since plugins are called one after the other? But it could happen
        # if a plugin forgets to pop items from the package list... The problem is that we
        # can't know which plugin did what, so we could only say "something went wrong"
        # and can't point to which plugin is at fault.

        # Remove empty groups
        _packageGroups = [group for group in _packageGroups if group]

        # Add packages that were not grouped.
        _packageGroups += [
            rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo](tuple([package]))
            for package in packages
        ]

        # TODO: Should we postpone downloading to the last minute if we can?
        _LOG.info("[bold]Downloading...")

        packageGroups: list[
            rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact]
        ] = rez_pip.download.downloadPackages(_packageGroups, wheelsDir)

        foundLocally = downloaded = 0
        for group in packageGroups:
            for package in group.packages:
                if not package.isDownloadRequired():
                    foundLocally += 1
                else:
                    downloaded += 1

        message = f"Downloaded {downloaded} wheels"
        if foundLocally:
            message += f"skipped {foundLocally} because they resolved to local files"

        _LOG.info(f"[bold]{message}")

        with rez_pip.utils.CONSOLE.status(
            f"[bold]Installing wheels into {installedWheelsDir!r}"
        ):
            for group in packageGroups:
                for package in group.packages:
                    _LOG.info(f"[bold]Installing {package.name!r} {package.path!r}")
                    targetPath = os.path.join(installedWheelsDir, package.name)
                    dist = rez_pip.install.installWheel(
                        package,
                        package.path,
                        targetPath,
                    )

                    rez_pip.install.cleanup(dist, targetPath)
                    rez_pip.patch.patch(dist, targetPath)

                    group.dists.append(dist)

        with rez_pip.utils.CONSOLE.status("[bold]Creating rez packages..."):
            for group in packageGroups:
                rez_pip.rez.createPackage(
                    group,
                    rez.version.Version(pythonVersion),
                    installedWheelsDir,
                    prefix=args.prefix,
                    release=args.release,
                )


def _debug(
    args: argparse.Namespace, console: rich.console.Console = rez_pip.utils.CONSOLE
) -> None:
    """Print debug information"""
    prefix = "  "
    console.print(
        f"[bold]rez-pip version[/]: {importlib_metadata.version(__package__)}"
    )

    console.print(f"[bold]rez version[/]: {importlib_metadata.version('rez')}")

    console.print(f"[bold]python version[/]: {sys.version}", highlight=False)
    console.print(f"[bold]python executable[/]: {sys.executable}", highlight=False)

    pip = args.pip or rez_pip.pip.getBundledPip()

    console.print(f"[bold]pip[/]: {pip}", highlight=False)

    completedProcess = subprocess.run(
        [sys.executable, pip, "--version"],
        stdout=subprocess.PIPE,
        text=True,
    )

    console.print(
        f"[bold]pip version[/]: {completedProcess.stdout.strip()}", highlight=False
    )

    completedProcess = subprocess.run(
        [sys.executable, pip, "config", "debug"],
        stdout=subprocess.PIPE,
        text=True,
    )

    opts = vars(args)

    # Get rid of non serializable values injected by rez.
    opts.pop("parser", None)
    opts.pop("func", None)
    opts.pop("formatter_class", None)

    console.print("[bold]rez-pip provided arguments[/]:")
    print(textwrap.indent(json.dumps(opts, indent=4), prefix))

    console.print(f"[bold]pip config debug[/]:", highlight=False)
    print(textwrap.indent(completedProcess.stdout.strip(), "  "))

    completedProcess = subprocess.run(
        [sys.executable, pip, "config", "list"],
        stdout=subprocess.PIPE,
        text=True,
    )

    console.print(f"[bold]pip config list[/]:", highlight=False)
    print(
        textwrap.indent(completedProcess.stdout.strip(), prefix)
        or f"{prefix}Returned nothing"
    )

    console.print(f"[bold]rez python packages[/]:", highlight=False)
    for pythonVersion, pythonExecutable in rez_pip.rez.getPythonExecutables(
        args.python_version
    ).items():
        print(textwrap.indent(f"{pythonExecutable} ({pythonVersion})", prefix))

    print()

    rich.print(
        rich.panel.Panel(
            rich.text.Text(
                "Please redact any sensitive information before giving this output to someone else!\n\n"
                "Don't remove things, only redact/replace. Look for IP addresses, domain names, passwords, etc.",
                justify="center",
            ),
            title="[bold red]WARNING!",
            expand=False,
            border_style="yellow",
        ),
        file=sys.stderr,
    )


def _printPlugins() -> None:
    table = rich.table.Table("Name", "Hooks", box=None)
    for plugin, hooks in rez_pip.plugins._getHookImplementations().items():
        table.add_row(plugin, ", ".join(hooks))
    rez_pip.utils.CONSOLE.print(table)


def run(
    args: argparse.Namespace | None = None, pipArgs: list[str] | None = None
) -> int:
    pipWorkArea = tempfile.mkdtemp(prefix="rez-pip-target")

    if args is None:
        args, pipArgs = _parseArgs(sys.argv[1:])

    assert args is not None
    assert pipArgs is not None

    if args.noop:
        print("Noop mode enabled")
        return 0

    # Initialize the plugin system
    rez_pip.plugins.getManager()

    if args.list_plugins:
        _printPlugins()
        return 0

    try:
        _validateArgs(args)

        handler = rich.logging.RichHandler(
            show_time=False,
            markup=True,
            show_path=False,
            console=rez_pip.utils.CONSOLE,
        )
        handler.setFormatter(logging.Formatter(fmt="%(message)s"))

        rootLogger = logging.getLogger("rez_pip")
        rootLogger.addHandler(handler)
        rootLogger.setLevel(args.log_level.upper())

        if args.debug_info:
            _debug(args)
            return 0

        _run(args, pipArgs, pipWorkArea)
        return 0
    except rez_pip.exceptions.RezPipError as exc:
        rez_pip.utils.CONSOLE.print(exc, soft_wrap=True)
        return 1
    finally:
        if not args.keep_tmp_dirs:
            _LOG.debug(f"Removing {pipWorkArea}")
            shutil.rmtree(pipWorkArea)



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

import sys


if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

__all__ = ["importlib_metadata"]



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pathlib
import unittest.mock

import pytest

import rez_pip.pip
import rez_pip.plugins
import rez_pip.exceptions

from . import utils


@pytest.fixture(scope="module", autouse=True)
def setupPluginManager():
    yield utils.initializePluginManager("shiboken6")


def fakePackage(name: str, **kwargs) -> unittest.mock.Mock:
    value = unittest.mock.MagicMock()
    value.configure_mock(name=name, **kwargs)
    return value


@pytest.mark.parametrize("package", [fakePackage("asd")])
def test_cleanup_noop(package, tmp_path: pathlib.Path):
    (tmp_path / "python" / "PySide6").mkdir(parents=True)

    rez_pip.plugins.getHook().cleanup(dist=package, path=tmp_path)

    assert (tmp_path / "python" / "PySide6").exists()


@pytest.mark.parametrize(
    "package",
    [
        fakePackage("shiboken6"),
        fakePackage("shiboken6_essentials"),
        fakePackage("ShIbOkEn6-AddoNs"),
    ],
)
def test_cleanup(package, tmp_path: pathlib.Path):
    (tmp_path / "python" / "PySide6").mkdir(parents=True)

    rez_pip.plugins.getHook().cleanup(dist=package, path=tmp_path)

    assert not (tmp_path / "python" / "shiboken6").exists()



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import re
import inspect
import argparse
import importlib

import docutils.nodes
import sphinx.transforms
import sphinx.util.nodes
import sphinx.application
import sphinx.ext.autodoc
import sphinx.util.docutils
import docutils.statemachine

import rez_pip.cli
import rez_pip.plugins

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "rez-pip"
copyright = "Contributors to the rez project"
author = "Rez Developers"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # first-party extensions
    "sphinx.ext.todo",
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    # Third-part
    "sphinx_inline_tabs",
    "myst_parser",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]


# -- Options for sphinx.ext.autosectionlabel ---------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autosectionlabel.html

autosectionlabel_prefix_document = True


# -- Options for the linkcheck builder ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-the-linkcheck-builder

linkcheck_allowed_redirects = {
    r"https://github.com/JeanChristopheMorinPerso/rez-pip/issues/\d+": r"https://github.com/JeanChristopheMorinPerso/rez-pip/pull/\d+"
}


# -- Options for sphinx.ext.intersphinx_mapping ------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "rez": ("https://rez.readthedocs.io/en/stable/", None),
}

# Force usage of :external:
# intersphinx_disabled_reftypes = ["*"]


# -- Options for sphinx.ext.autodoc ------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html

# autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_member_order = "bysource"

# -- Options for sphinx.ext.todo --------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html

todo_include_todos = True

# -- Custom ------------------------------------------------------------------
# Custom stuff


class ReplaceGHRefs(sphinx.transforms.SphinxTransform):
    default_priority = 750
    prefix = "https://github.com/JeanChristopheMorinPerso/rez-pip/issues"

    def apply(self):
        istext = lambda o: isinstance(o, docutils.nodes.Text)

        for node in self.document.traverse(istext):
            # Handle issue/PR references like (#123)
            match = re.match(r".*\((\#(\d+))\)\.?$", str(node))
            if match:
                newtext = docutils.nodes.Text(str(node)[: match.start(1)])
                newreference = docutils.nodes.reference(
                    "", match.group(1), refuri=f"{self.prefix}/{match.group(2)}"
                )
                trailingtext = docutils.nodes.Text(str(node)[match.end(1) :])
                node.parent.replace(node, newtext)
                node.parent.insert(node.parent.index(newtext) + 1, newreference)
                node.parent.insert(node.parent.index(newtext) + 2, trailingtext)
                continue

            # Handle user tags like @username
            match = re.search(r"(@([a-zA-Z0-9-]+(?:\[bot\])?))", str(node))
            if match:
                newtext = docutils.nodes.Text(str(node)[: match.start(1)])
                newreference = docutils.nodes.reference(
                    "", match.group(1), refuri=f"https://github.com/{match.group(2)}"
                )
                trailingtext = docutils.nodes.Text(str(node)[match.end(1) :])
                node.parent.replace(node, newtext)
                node.parent.insert(node.parent.index(newtext) + 1, newreference)
                node.parent.insert(node.parent.index(newtext) + 2, trailingtext)

        return


class RezAutoArgparseDirective(sphinx.util.docutils.SphinxDirective):
    """
    Special rez-autoargparse directive. This is quite similar to "autosummary" in some ways.
    """

    required_arguments = 0
    optional_arguments = 0

    def run(self) -> list[docutils.nodes.Node]:
        # Create the node.
        node = docutils.nodes.section()
        node.document = self.state.document

        rst = docutils.statemachine.ViewList()

        # Add rezconfig as a dependency to the current document. The document
        # will be rebuilt if rezconfig changes.
        self.env.note_dependency(rez_pip.cli.__file__)
        self.env.note_dependency(__file__)

        path, lineNumber = self.get_source_info()

        parser = rez_pip.cli._createParser()

        full_cmd = parser.prog.replace(" ", "-").replace("rez-pip", "rez pip2")

        # Title
        document = [f".. _{full_cmd}:"]
        document.append("")
        document.append(f"{'='*len(full_cmd)}")
        document.append(f"{full_cmd}")
        document.append(f"{'='*len(full_cmd)}")
        document.append("")

        document.append(f".. program:: {full_cmd}")
        document.append("")
        document.append("Usage")
        document.append("=====")
        document.append("")
        document.append(".. code-block:: text")
        document.append("")
        for line in parser.format_usage()[7:].split("\n"):
            # TODO: This replace is dirty... How could we more cleany
            # get the rez pip command as a plugin?
            document.append(f"   {line}".replace("rez-pip", "rez pip2"))
        document.append("")

        document.append("Description")
        document.append("===========")
        document.extend(parser.description.split("\n"))

        for group in parser._action_groups:
            if not group._group_actions:
                continue

            document.append("")
            title = group.title.capitalize()
            document.append(title)
            document.append("=" * len(title))
            document.append("")

            for action in group._group_actions:
                if isinstance(action, argparse._HelpAction):
                    continue

                # Quote default values for string/None types
                default = action.default
                if (
                    action.default not in ["", None, True, False]
                    and action.type in [None, str]
                    and isinstance(action.default, str)
                ):
                    default = f'"{default}"'

                # fill in any formatters, like %(default)s
                format_dict = dict(vars(action), prog=parser.prog, default=default)
                format_dict["default"] = default
                help_str = action.help or ""  # Ensure we don't print None
                try:
                    help_str = help_str % format_dict
                except Exception:
                    pass

                if help_str == argparse.SUPPRESS:
                    continue

                # Avoid Sphinx warnings.
                help_str = help_str.replace("*", "\\*")
                # Replace everything that looks like an argument with an option directive.
                help_str = re.sub(
                    r"(?<!\w)-[a-zA-Z](?=\s|\/|\)|\.?$)|(?<!\w)--[a-zA-Z-0-9]+(?=\s|\/|\)|\.?$)",
                    r":option:`\g<0>`",
                    help_str,
                )
                help_str = help_str.replace("--", "\\--")

                # Add links to rez docs for known settings.
                help_str = re.sub(
                    r"(.* \(default: configured )([a-zA-Z_]+)(.*)$",
                    r"\g<1> :external:data:`\g<2>`\g<3>",
                    help_str,
                )

                # Options have the option_strings set, positional arguments don't
                name = action.option_strings
                if name == []:
                    if action.metavar is None:
                        name = [action.dest]
                    else:
                        name = [action.metavar]

                # Skip lines for subcommands
                if name == [argparse.SUPPRESS]:
                    continue

                metavar = action.metavar if action.metavar else ""
                document.append(f".. option:: {', '.join(name)} {metavar.lower()}")
                document.append("")
                document.append(f"   {help_str}")
                if action.choices:
                    document.append("")
                    document.append(f"   Choices: {', '.join(action.choices)}")
                document.append("")

        document = "\n".join(document)

        # Add each line to the view list.
        for index, line in enumerate(document.split("\n")):
            # Note to future people that will look at this.
            # "line" has to be a single line! It can't be a line like "this\nthat".
            rst.append(line, path, lineNumber + index)

        # Finally, convert the rst into the appropriate docutils/sphinx nodes.
        sphinx.util.nodes.nested_parse_with_titles(self.state, rst, node)

        # Return the generated nodes.
        return node.children


class RezAutoPlugins(sphinx.util.docutils.SphinxDirective):
    """
    Special rez-pip-autoplugins directive. This is quite similar to "autosummary" in some ways.
    """

    required_arguments = 0
    optional_arguments = 0

    def run(self) -> list[docutils.nodes.Node]:
        # Create the node.
        node = docutils.nodes.section()
        node.document = self.state.document

        rst = docutils.statemachine.ViewList()

        # Add rezconfig as a dependency to the current document. The document
        # will be rebuilt if rezconfig changes.
        self.env.note_dependency(rez_pip.plugins.__file__)
        self.env.note_dependency(__file__)

        path, lineNumber = self.get_source_info()

        document = []
        for plugin, hooks in rez_pip.plugins._getHookImplementations().items():
            hooks = [f":func:`{hook}`" for hook in hooks]
            document.append(f"* {plugin.split('.')[-1]}: {', '.join(hooks)}")

        document = "\n".join(document)

        # Add each line to the view list.
        for index, line in enumerate(document.split("\n")):
            # Note to future people that will look at this.
            # "line" has to be a single line! It can't be a line like "this\nthat".
            rst.append(line, path, lineNumber + index)

        # Finally, convert the rst into the appropriate docutils/sphinx nodes.
        sphinx.util.nodes.nested_parse_with_titles(self.state, rst, node)

        # Return the generated nodes.
        return node.children


class RezPipAutoPluginHooks(sphinx.util.docutils.SphinxDirective):
    """
    Special rez-pip-autopluginhooks directive. This is quite similar to "autosummary" in some ways.
    """

    required_arguments = 1
    optional_arguments = 0

    def run(self) -> list[docutils.nodes.Node]:
        # Create the node.
        node = docutils.nodes.section()
        node.document = self.state.document

        rst = docutils.statemachine.ViewList()

        # Add rezconfig as a dependency to the current document. The document
        # will be rebuilt if rezconfig changes.
        self.env.note_dependency(rez_pip.plugins.__file__)
        self.env.note_dependency(__file__)

        path, lineNumber = self.get_source_info()

        fullyQualifiedClassName = self.arguments[0]
        module, klassname = fullyQualifiedClassName.rsplit(".", 1)

        mod = importlib.import_module(module)
        klass = getattr(mod, klassname)

        methods = [
            method
            for method in inspect.getmembers(klass, predicate=inspect.isfunction)
            if not method[0].startswith("_")
        ]

        document = []
        for method in sorted(methods, key=lambda x: x[1].__code__.co_firstlineno):
            document.append(f".. autohook:: {module}.{klassname}.{method[0]}")

        document = "\n".join(document)

        # Add each line to the view list.
        for index, line in enumerate(document.split("\n")):
            # Note to future people that will look at this.
            # "line" has to be a single line! It can't be a line like "this\nthat".
            rst.append(line, path, lineNumber + index)

        # Finally, convert the rst into the appropriate docutils/sphinx nodes.
        sphinx.util.nodes.nested_parse_with_titles(self.state, rst, node)

        # Return the generated nodes.
        return node.children


def autodoc_process_signature(
    app: sphinx.application.Sphinx,
    what: str,
    name: str,
    obj,
    options: dict,
    signature: str,
    return_annotation,
):
    signature = signature.replace(
        "rez_pip.compat.importlib_metadata", "~importlib.metadata"
    )

    return signature, return_annotation


class HookDocumenter(sphinx.ext.autodoc.FunctionDocumenter):
    """
    Custom autohook directive to document our hooks.
    It allows us to easily document the hooks from the rez_pip.plugins.PluginSpec
    class without exposing the class and module name.
    """

    objtype = "hook"  # auto + hook
    directivetype = "function"  # generated reST directive

    def format_signature(self, **kwargs) -> str:
        """
        Format the signature and remove self. We really don't want to expose
        the class and module name or the fact that we are documenting methods.
        """
        sig = super().format_signature(**kwargs)
        sig = re.sub(r"\(self(,\s)?", "(", sig)

        # Also force short names for our own types
        sig = sig.replace("rez_pip.", "~rez_pip.")

        return sig

    def add_directive_header(self, sig):
        modname = self.modname
        # Hacky, but it does the job. This should remove the module name from the directive
        # created by autodoc.
        self.modname = ""

        data = super().add_directive_header(sig)

        # We need to restore it because autodoc does lots of things with the module name.
        self.modname = modname
        return data


def setup(app: sphinx.application.Sphinx):
    app.add_directive("rez-autoargparse", RezAutoArgparseDirective)
    app.add_directive("rez-pip-autoplugins", RezAutoPlugins)
    app.add_directive("rez-pip-autopluginhooks", RezPipAutoPluginHooks)
    app.add_transform(ReplaceGHRefs)

    app.connect("autodoc-process-signature", autodoc_process_signature)
    app.add_autodocumenter(HookDocumenter)



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""
Code that takes care of installing (extracting) wheels.
"""

from __future__ import annotations

import io
import os
import re
import sys
import shutil
import typing
import logging
import pathlib
import zipfile
import sysconfig
import collections.abc

import rez_pip.exceptions

if typing.TYPE_CHECKING:
    from typing import Literal

import installer
import installer.utils
import installer.records
import installer.scripts
import installer.sources
import installer.destinations

import rez_pip.pip
import rez_pip.plugins
import rez_pip.exceptions
from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    LauncherKind = Literal["posix", "win-ia32", "win-amd64", "win-arm", "win-arm64"]
    ScriptSection = Literal["console", "gui"]


class CleanupError(rez_pip.exceptions.RezPipError):
    """
    Raised when a cleanup operation fails.
    """


def isWheelPure(dist: importlib_metadata.Distribution) -> bool:
    # dist.files should never be empty, but assert to silence mypy.
    assert dist.files is not None

    path = next(
        f
        for f in dist.files
        if os.fspath(f.locate()).endswith(os.path.join(".dist-info", "WHEEL"))
    )
    with open(path.locate()) as fd:
        metadata = installer.utils.parse_metadata_file(fd.read())
    return typing.cast(str, metadata["Root-Is-Purelib"]) == "true"


# Taken from https://github.com/pypa/installer/blob/main/src/installer/__main__.py#L49
def getSchemeDict(name: str, target: str) -> dict[str, str]:
    vars = {}
    vars["base"] = vars["platbase"] = installed_base = target

    schemeDict = sysconfig.get_paths(vars=vars)
    # calculate 'headers' path, not currently in sysconfig - see
    # https://bugs.python.org/issue44445. This is based on what distutils does.
    # TODO: figure out original vs normalised distribution names
    schemeDict["headers"] = os.path.join(
        sysconfig.get_path("include", vars={"installed_base": installed_base}),
        name,
    )

    if target:  # In practice this will always be set.
        schemeDict["purelib"] = os.path.join(target, "python")
        schemeDict["platlib"] = os.path.join(target, "python")
        schemeDict["headers"] = os.path.join(target, "headers", name)
        schemeDict["scripts"] = os.path.join(target, "scripts")
        # Potentiall handle data?
    return schemeDict


def installWheel(
    package: rez_pip.pip.PackageInfo,
    wheelPath: str,
    targetPath: str,
) -> importlib_metadata.Distribution:
    # TODO: Technically, target should be optional. We will always want to install in "pip install --target"
    #       mode. So right now it's a CLI option for debugging purposes.

    destination = CustomWheelDestination(
        getSchemeDict(package.name, targetPath),
        # TODO: Use Python from rez package, or simply use "/usr/bin/env python"?
        interpreter=sys.executable,
        script_kind=installer.utils.get_launcher_kind(),
    )

    _LOG.debug(f"Installing {wheelPath} into {targetPath!r}")
    with installer.sources.WheelFile.open(pathlib.Path(wheelPath)) as source:
        installer.install(
            source=source,
            destination=destination,
            # Additional metadata that is generated by the installation tool.
            additional_metadata={
                "INSTALLER": f"rez-pip {importlib_metadata.version(__package__)}".encode(
                    "utf-8"
                ),
            },
        )

    targetPathPython = os.path.join(targetPath, "python")

    # That's kind of dirty, but using any other method returns inconsistent results.
    # For example, importlib.metadata.Distribution.discover(path=['/path']) sometimes
    # won't find the freshly intalled package, even if it exists and everything.
    items = [
        item
        for item in os.listdir(targetPathPython)
        if item.endswith(".dist-info")
        and os.path.isdir(os.path.join(targetPathPython, item))
    ]

    if len(items) == 0:
        raise rez_pip.exceptions.RezPipError(
            f"Could not find a dist-info folder for {package.name!r} in {targetPathPython!r}"
        )

    elif len(items) > 1:
        raise rez_pip.exceptions.RezPipError(
            f"Expected only one dist-info folders for {package.name!r} in {targetPathPython!r}, but found {len(items)}: {items}"
        )

    dist = importlib_metadata.Distribution.at(os.path.join(targetPathPython, items[0]))

    return dist


# TODO: Document where this code comes from.
class CustomWheelDestination(installer.destinations.SchemeDictionaryDestination):
    # Exactly the same as SchemeDictionaryDestination, but uses our custom Script class.
    def write_script(
        self, name: str, module: str, attr: str, section: ScriptSection
    ) -> installer.records.RecordEntry:
        """Write a script to invoke an entrypoint.
        :param name: name of the script
        :param module: module path, to load the entry point from
        :param attr: final attribute access, for the entry point
        :param section: Denotes the "entry point section" where this was specified.
            Valid values are ``"gui"`` and ``"console"``.
        :type section: str
        - Generates a launcher using :any:`Script.generate`.
        - Writes to the "scripts" scheme.
        - Uses :py:meth:`SchemeDictionaryDestination.write_to_fs` for the
          filesystem interaction.
        """
        script = Script(name, module, attr, section)
        script_name, data = script.generate(self.interpreter, self.script_kind)

        with io.BytesIO(data) as stream:
            entry = self.write_to_fs(
                installer.utils.Scheme("scripts"),
                script_name,
                stream,
                is_executable=True,
            )

            path = self._path_with_destdir(
                installer.utils.Scheme("scripts"), script_name
            )
            mode = os.stat(path).st_mode
            mode |= (mode & 0o444) >> 2
            os.chmod(path, mode)

            return entry


_SCRIPT_TEMPLATE = """\
# -*- coding: utf-8 -*-
import re
import sys
from {module} import {import_name}
if __name__ == "__main__":
    sys.argv[0] = re.sub(r"(-script\\.pyw|\\.exe)?$", "", sys.argv[0])
    sys.exit({func_path}())
"""


# TODO: Document where this code comes from.
class Script(installer.scripts.Script):
    def generate(self, executable: str, kind: LauncherKind) -> tuple[str, bytes]:
        """Generate a launcher for this script.
        :param executable: Path to the executable to invoke.
        :param kind: Which launcher template should be used.
            Valid values are ``"posix"``, ``"win-ia32"``, ``"win-amd64"`` and
            ``"win-arm"``.
        :type kind: str
        :raises InvalidScript: if no appropriate template is available.
        :return: The name and contents of the launcher file.
        """
        launcher = self._get_launcher_data(kind)
        # shebang = _build_shebang(executable, forlauncher=bool(launcher))
        # TODO: Should we instead just pass that value to WheelDestination?
        shebang = b"#!/usr/bin/env python"
        code = _SCRIPT_TEMPLATE.format(
            module=self.module,
            import_name=self.attr.split(".")[0],
            func_path=self.attr,
        ).encode("utf-8")

        if launcher is None:
            return (self.name, shebang + b"\n" + code)

        stream = io.BytesIO()
        with zipfile.ZipFile(stream, "w") as zf:
            zf.writestr("__main__.py", code)
        name = f"{self.name}.exe"
        data = launcher + shebang + b"\n" + stream.getvalue()
        return (name, data)


def cleanup(dist: importlib_metadata.Distribution, path: str) -> None:
    """
    Run cleanup hooks.

    Note that this lives in install because the cleanups
    are made on the installs (the wheel install). We could move this somewhere
    else but it's not clear where.
    """
    actionsGroups: collections.abc.Sequence[
        collections.abc.Sequence[rez_pip.plugins.CleanupAction]
    ] = rez_pip.plugins.getHook().cleanup(
        dist=dist, path=path
    )  # type: ignore[assignment]

    # Flatten
    actions: list[rez_pip.plugins.CleanupAction] = [
        action for group in actionsGroups for action in group
    ]

    recordEntriesToRemove = []

    for action in actions:
        if not action.path.startswith(path):
            # Security measure. Only perform operations on
            # paths that are within the install path.
            raise CleanupError(
                f"Typing to {action.op} {action.path!r} which is outside of {path!r}"
            )

        if action.op == "remove":
            if not os.path.exists(action.path):
                continue

            _LOG.info(f"Removing {action.path!r}")
            if os.path.isdir(action.path):
                shutil.rmtree(action.path)
            else:
                os.remove(action.path)

            recordEntriesToRemove.append(
                os.path.normpath(os.path.relpath(action.path, path)).replace("\\", "/")
            )
        else:
            raise CleanupError(f"Unknown action: {action.op}")

    if recordEntriesToRemove:
        deleteEntryFromRecord(dist, path, recordEntriesToRemove)


def deleteEntryFromRecord(
    dist: importlib_metadata.Distribution, path: str, entries: list[str]
) -> None:
    """
    Delete an entry from the record file.

    This code is not great. I feel like updating the RECORD file should
    be simpler. Which means that we miht need to refactor things a bit.
    """
    items = [
        os.fspath(item)
        for item in dist.files
        if re.search(r"[a-zA-Z0-9._+]+\.dist-info/RECORD", os.fspath(item))
    ]

    if not items:
        raise CleanupError(f"RECORD file not found for {dist.name!r}")

    recordFilePathRel = items[0]
    recordFilePath = os.path.join(path, "python", recordFilePathRel)

    with open(recordFilePath) as f:
        lines = f.readlines()

    schemesRaw = getSchemeDict(dist.name, path)
    schemes = {
        key: os.path.relpath(value, path)
        for key, value in schemesRaw.items()
        if value.startswith(path)
    }

    # Format the entries to match the record file. This is important
    # because when we install the files, we use a custom scheme.
    # For example, we have to trim "python/" or "scripts/".
    for index, entry in enumerate(entries):
        for schemePath in schemes.values():
            if entry.startswith(schemePath):
                _LOG.debug(f"Stripping {schemePath!r}/ from {entry!r}")
                entries[index] = entry.lstrip(schemePath + "/")
                # Break on first match
                break

    for index, elements in enumerate(installer.records.parse_record_file(lines)):
        if elements[0] in entries:
            lines.pop(index)

    with open(recordFilePath, "w") as f:
        for line in lines:
            f.write(line)



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""PySide6 plugin."""

from __future__ import annotations

import os
import typing
import logging
import platform

import packaging.utils
import packaging.version
import packaging.specifiers
import packaging.requirements

import rez_pip.pip
import rez_pip.patch
import rez_pip.plugins
import rez_pip.exceptions

if typing.TYPE_CHECKING:
    from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger(__name__)


@rez_pip.plugins.hookimpl
def prePipResolve(
    packages: tuple[str],
) -> None:
    """
    PySide6 was initially a single package that had shiboken as a dependency.
    Starting from 6.3.0, the package was spit in 3, PySide6, PySide6-Essentials and
    PySide6-Addons.

    So we need to intercept what the user installs and install all 3 packages together.
    Not doing that would result in a broken install (eventually).
    """
    pyside6Seen = False
    variantsSeens = []

    for package in packages:
        req = packaging.requirements.Requirement(package)
        name = packaging.utils.canonicalize_name(req.name)

        if name == "pyside6":
            pyside6Seen = True
        elif name in ["pyside6-essentials", "pyside6-addons"]:
            variantsSeens.append(req.name)

    if variantsSeens and not pyside6Seen:
        variants = " and ".join(variantsSeens)
        verb = "was" if len(variantsSeens) == 1 else "were"
        raise rez_pip.exceptions.RezPipError(
            f"{variants} {verb} requested but PySide6 was not. You must explicitly request PySide6 in addition to {variants}."
        )


@rez_pip.plugins.hookimpl
def postPipResolve(packages: tuple[rez_pip.pip.PackageInfo]) -> None:
    """
    This hook is implemented out of extra caution. We really don't want PySide6-Addons
    or PySide6-Essentials to be installed without PySide6.

    In this case, we cover cases where a user requests a package X and that package
    depends on PySide6-Addons or PySide6-Essentials.
    """
    pyside6Seen = False
    variantsSeens = []

    for package in packages:
        name = packaging.utils.canonicalize_name(package.name)
        if name == "pyside6":
            pyside6Seen = True
        elif name in ["pyside6-essentials", "pyside6-addons"]:
            variantsSeens.append(package.name)

    if variantsSeens and not pyside6Seen:
        variants = " and ".join(variantsSeens)
        verb = "is" if len(variantsSeens) == 1 else "are"
        raise rez_pip.exceptions.RezPipError(
            f"{variants} {verb} part of the resolved packages but PySide6 was not. Dependencies and or you must explicitly request PySide6 in addition to {variants}."
        )


@rez_pip.plugins.hookimpl
def groupPackages(
    packages: list[rez_pip.pip.PackageInfo],
) -> list[rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo]]:
    data = []
    for package in packages[:]:
        if packaging.utils.canonicalize_name(package.name) in [
            "pyside6",
            "pyside6-addons",
            "pyside6-essentials",
        ]:
            data.append(package)
            packages.remove(package)

    return [rez_pip.pip.PackageGroup[rez_pip.pip.PackageInfo](tuple(data))]


@rez_pip.plugins.hookimpl
def patches(dist: importlib_metadata.Distribution, path: str) -> list[str]:
    if dist.name != "PySide6" or platform.system() != "Windows":
        return []

    patches: list[str] = []

    # To generate the patches:
    # 1. run srcipts/get_pyside6_files.py
    # 2. Look at the output and get the different diffs.
    # 3. Generate a patch for each version range where there is a diff:
    # 3.1 cp patches/data/6.3.0/__init__.py patches/data/6.3.0/__init__.py.new
    # 3.2 vim patches/data/6.3.0/__init__.py.new
    # 3.3 diff --unified patches/data/6.3.0/__init__.py patches/data/6.3.0/__init__.py.new > src/rez_pip/data/patches/pyside6_6_3_0_win_dll_path.patch
    # 3.4 Edit src/rez_pip/data/patches/pyside6_6_3_0_win_dll_path.patch to make sure the paths
    #     are good in the patch header. Also make sure CRLF (line ending) is kept as-is.
    #
    # I think that it's important that the paths in the look like "python/PySide6/..."
    # so that it matches the file layout...
    version = packaging.version.Version(dist.version)
    if version in packaging.specifiers.SpecifierSet(">=6.0.0,<6.1.0"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_0_0_win_dll_path.patch"
            )
        )

    elif version in packaging.specifiers.SpecifierSet(">=6.1.0,<6.2.4"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_1_0_win_dll_path.patch"
            )
        )

    elif version in packaging.specifiers.SpecifierSet(">=6.2.4,<6.3.0"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_2_4_win_dll_path.patch"
            )
        )

    elif version in packaging.specifiers.SpecifierSet(">=6.3.0,<6.7.3"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_3_0_win_dll_path.patch"
            )
        )

    elif version in packaging.specifiers.SpecifierSet(">=6.7.3,<6.8.1"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_7_3_win_dll_path.patch"
            )
        )
    elif version in packaging.specifiers.SpecifierSet(">=6.8.1"):
        patches.append(
            os.path.join(
                rez_pip.patch.getBuiltinPatchesDir(), "pyside6_6_8_1_win_dll_path.patch"
            )
        )
    return patches


@rez_pip.plugins.hookimpl
def cleanup(
    dist: importlib_metadata.Distribution, path: str
) -> list[rez_pip.plugins.CleanupAction]:
    actions: list[rez_pip.plugins.CleanupAction] = []

    if packaging.utils.canonicalize_name(dist.name) not in [
        "pyside6",
        "pyside6-addons",
        "pyside6-essentials",
    ]:
        return actions

    # Remove shiboken6 from PySide6 packages...
    # PySide6 >=6.3, <6.6.2 were shipping some shiboken6 folders by mistake.
    # Not removing these extra folders would stop python from being able to import
    # the correct shiboken (that lives in a separate rez package).
    actions.extend(
        [
            rez_pip.plugins.CleanupAction(
                "remove", os.path.join(path, "python", "shiboken6")
            ),
            rez_pip.plugins.CleanupAction(
                "remove", os.path.join(path, "python", "shiboken6_generator")
            ),
        ]
    )

    if packaging.utils.canonicalize_name(dist.name) in [
        "pyside6-addons",
        "pyside6-essentials",
    ]:
        # Because we patch __init__.py, we need to make sure that
        # PySide6-Addons and PySide6-Essentials' _init__.py won't
        # overrite our patched __init__.py.
        actions.append(
            rez_pip.plugins.CleanupAction(
                "remove", os.path.join(path, "python", "PySide6", "__init__.py")
            )
        )

    return actions



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""Plugin system."""

from __future__ import annotations

import typing
import logging
import pkgutil
import functools
import importlib
import dataclasses
import collections.abc

import pluggy
import rez.package_maker

if typing.TYPE_CHECKING:
    import rez_pip.pip
    import rez_pip.compat

__all__ = [
    "hookimpl",
]


def __dir__() -> list[str]:
    return __all__


_LOG = logging.getLogger(__name__)

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])
hookspec = typing.cast(typing.Callable[[F], F], pluggy.HookspecMarker("rez-pip"))
hookimpl = typing.cast(typing.Callable[[F], F], pluggy.HookimplMarker("rez-pip"))


@dataclasses.dataclass(frozen=True)
class CleanupAction:
    """
    Cleanup hook action. If you want to do any cleanup from a cleanup hook
    you need to return this from the :func:`cleanup` hook.
    """

    #: Operation to perform.
    op: typing.Literal["remove"]

    #: Path on which to perform the operation.
    path: str


class PluginSpec:
    @hookspec
    def prePipResolve(
        self,
        packages: tuple[str, ...],  # Immutable
        requirements: tuple[str, ...],  # Immutable
    ) -> None:
        """
        The pre-pip resolve hook allows a plugin to run some checks *before* resolving the
        requested packages using pip. The hook **must** not modify the content of the
        arguments passed to it.

        Some use cases are allowing or disallowing the installation of some packages.

        :param packages: List of packages requested by the user.
        :param requirements: List of `requirements files <https://pip.pypa.io/en/stable/reference/requirements-file-format/#requirements-file-format>`_ if any.
        """
        ...

    @hookspec
    def postPipResolve(
        self,
        packages: tuple[rez_pip.pip.PackageInfo, ...],  # Immutable
    ) -> None:
        """
        The post-pip resolve hook allows a plugin to run some checks *after* resolving the
        requested packages using pip. The hook **must** not modify the content of the
        arguments passed to it.

        Some use cases are allowing or disallowing the installation of some packages.

        :param packages: List of resolved packages.
        """
        ...

    @hookspec
    def groupPackages(  # type: ignore[empty-body]
        self,
        packages: collections.abc.MutableSequence[rez_pip.pip.PackageInfo],
    ) -> collections.abc.Sequence[
        rez_pip.pip.PackageGroup[rez_pip.pip.DownloadedArtifact]
    ]:
        """
        Merge packages into groups of packages. The name and version of the first package
        in the group will be used as the name and version for the rez package.

        The hook **must** pop grouped packages out of the "packages" variable.

        :param packages: List of resolved packages.
        :returns: A list of package groups.
        """
        ...

    @hookspec
    def cleanup(  # type: ignore[empty-body]
        self, dist: rez_pip.compat.importlib_metadata.Distribution, path: str
    ) -> collections.abc.Sequence[CleanupAction]:
        """
        Cleanup a package post-installation. Do not delete any files/directories from this hook.
        Return the list of actions you want to perform and let rez-pip perform them.

        :param dist: Python distribution.
        :param path: Root path of the rez variant.
        """
        ...

    @hookspec
    def patches(  # type: ignore[empty-body]
        self, dist: rez_pip.compat.importlib_metadata.Distribution, path: str
    ) -> collections.abc.Sequence[str]:
        """
        Provide paths to patches to be applied on the source code of a package.

        :param dist: Python distribution.
        :param path: Root path of the installed content.
        """
        # TODO: This will alter files (obviously) and change their hashes.
        # This could be a problem to verify the integrity of the package.
        # https://packaging.python.org/en/latest/specifications/recording-installed-packages/#the-record-file
        ...

    @hookspec
    def metadata(self, package: rez.package_maker.PackageMaker) -> None:
        """
        Modify/inject metadata in the rez package. The plugin is expected to modify
        "package" in place.

        :param package: An insatnce of :class:`rez.package_maker.PackageMaker`.
        """
        ...


def before(
    hookName: str,
    hookImpls: collections.abc.Sequence[pluggy.HookImpl],
    kwargs: collections.abc.Mapping[str, typing.Any],
) -> None:
    """Function that will be called before each hook."""
    _LOG.debug("Calling the %r hooks", hookName)


def after(
    outcome: pluggy.Result[typing.Any],
    hookName: str,
    hookImpls: collections.abc.Sequence[pluggy.HookImpl],
    kwargs: collections.abc.Mapping[str, typing.Any],
) -> None:
    """Function that will be called after each hook."""
    _LOG.debug("Called the %r hooks", hookName)


@functools.lru_cache
def getManager() -> pluggy.PluginManager:
    """
    Returns the plugin manager. The return value will be cached on first call
    and the cached value will be return in subsequent calls.
    """
    manager = pluggy.PluginManager("rez-pip")
    if _LOG.getEffectiveLevel() <= logging.DEBUG:
        manager.trace.root.setwriter(print)
        manager.enable_tracing()

    manager.add_hookspecs(PluginSpec)

    # Register the builtin plugins
    for module in pkgutil.iter_modules(__path__):
        manager.register(
            importlib.import_module(f"rez_pip.plugins.{module.name}"),
            name=f"rez_pip.{module.name}",
        )

    manager.load_setuptools_entrypoints("rez-pip")

    manager.add_hookcall_monitoring(before, after)
    return manager


def getHook() -> PluginSpec:
    """
    Returns the hook attribute from the manager. This is allows
    to have type hints at the caller sites.

    Inspired by https://stackoverflow.com/a/54695761.
    """
    manager = getManager()
    return typing.cast(PluginSpec, manager.hook)


def _getHookImplementations() -> dict[str, list[str]]:
    manager = getManager()

    implementations = {}
    for name, plugin in manager.list_name_plugin():
        hookcallers = manager.get_hookcallers(plugin)

        # hookcallers will never be None because we get the names from list_name_plugin.
        # But it silences mypy.
        assert hookcallers is not None

        implementations[name] = [caller.name for caller in hookcallers]
    return implementations



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

"""shiboken6 plugin."""

from __future__ import annotations

import os
import typing
import logging

import packaging.utils

import rez_pip.plugins

if typing.TYPE_CHECKING:
    from rez_pip.compat import importlib_metadata

_LOG = logging.getLogger(__name__)


@rez_pip.plugins.hookimpl
def cleanup(
    dist: importlib_metadata.Distribution, path: str
) -> list[rez_pip.plugins.CleanupAction]:
    if packaging.utils.canonicalize_name(dist.name) != "shiboken6":
        return []

    # Remove PySide6 from shiboken6 packages...
    # shiboken6 >=6.3, <6.6.2 were shipping some PySide6 folders by mistake.
    # Not removing these extra folders would stop python from being able to import
    # the correct PySide6 (that lives in a separate rez package).
    return [
        rez_pip.plugins.CleanupAction("remove", os.path.join(path, "python", "PySide6"))
    ]



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import ast
import sys
import bisect
import typing
import difflib
import zipfile
import tempfile
import itertools
import contextlib
import subprocess

import requests
import requests.models
import packaging.utils


# Token from https://github.com/pypa/pip/blob/bc553db53c264abe3bb63c6bcd6fc6f303c6f6e3/src/pip/_internal/network/lazy_wheel.py
class LazyZipOverHTTP:
    """File-like object mapped to a ZIP file over HTTP.

    This uses HTTP range requests to lazily fetch the file's content,
    which is supposed to be fed to ZipFile.  If such requests are not
    supported by the server, raise HTTPRangeRequestUnsupported
    during initialization.
    """

    def __init__(
        self,
        url: str,
        session: requests.Session,
        chunk_size: int = requests.models.CONTENT_CHUNK_SIZE,
    ) -> None:
        head = session.head(url, headers={"Accept-Encoding": "identity"})
        head.raise_for_status()
        assert head.status_code == 200
        self._session, self._url, self._chunk_size = session, url, chunk_size
        self._length = int(head.headers["Content-Length"])
        self._file = tempfile.NamedTemporaryFile()
        self.truncate(self._length)
        self._left: list[int] = []
        self._right: list[int] = []
        if "bytes" not in head.headers.get("Accept-Ranges", "none"):
            raise ValueError("range request is not supported")
        self._check_zip()

    @property
    def mode(self) -> str:
        """Opening mode, which is always rb."""
        return "rb"

    @property
    def name(self) -> str:
        """Path to the underlying file."""
        return self._file.name

    def seekable(self) -> bool:
        """Return whether random access is supported, which is True."""
        return True

    def close(self) -> None:
        """Close the file."""
        self._file.close()

    @property
    def closed(self) -> bool:
        """Whether the file is closed."""
        return self._file.closed

    def read(self, size: int = -1) -> bytes:
        """Read up to size bytes from the object and return them.

        As a convenience, if size is unspecified or -1,
        all bytes until EOF are returned.  Fewer than
        size bytes may be returned if EOF is reached.
        """
        download_size = max(size, self._chunk_size)
        start, length = self.tell(), self._length
        stop = length if size < 0 else min(start + download_size, length)
        start = max(0, stop - download_size)
        self._download(start, stop - 1)
        return self._file.read(size)

    def readable(self) -> bool:
        """Return whether the file is readable, which is True."""
        return True

    def seek(self, offset: int, whence: int = 0) -> int:
        """Change stream position and return the new absolute position.

        Seek to offset relative position indicated by whence:
        * 0: Start of stream (the default).  pos should be >= 0;
        * 1: Current position - pos may be negative;
        * 2: End of stream - pos usually negative.
        """
        return self._file.seek(offset, whence)

    def tell(self) -> int:
        """Return the current position."""
        return self._file.tell()

    def truncate(self, size: int | None = None) -> int:
        """Resize the stream to the given size in bytes.

        If size is unspecified resize to the current position.
        The current stream position isn't changed.

        Return the new file size.
        """
        return self._file.truncate(size)

    def writable(self) -> bool:
        """Return False."""
        return False

    def __enter__(self) -> LazyZipOverHTTP:
        self._file.__enter__()
        return self

    def __exit__(self, *exc: Any) -> None:
        self._file.__exit__(*exc)

    @contextlib.contextmanager
    def _stay(self) -> typing.Generator[None]:
        """Return a context manager keeping the position.

        At the end of the block, seek back to original position.
        """
        pos = self.tell()
        try:
            yield
        finally:
            self.seek(pos)

    def _check_zip(self) -> None:
        """Check and download until the file is a valid ZIP."""
        end = self._length - 1
        for start in reversed(range(0, end, self._chunk_size)):
            self._download(start, end)
            with self._stay():
                try:
                    # For read-only ZIP files, ZipFile only needs
                    # methods read, seek, seekable and tell.
                    zipfile.ZipFile(self)
                except zipfile.BadZipFile:
                    pass
                else:
                    break

    def _stream_response(
        self,
        start: int,
        end: int,
        base_headers: dict[str, str] = {"Accept-Encoding": "identity"},
    ) -> requests.Response:
        """Return HTTP response to a range request from start to end."""
        headers = base_headers.copy()
        headers["Range"] = f"bytes={start}-{end}"
        # TODO: Get range requests to be correctly cached
        headers["Cache-Control"] = "no-cache"
        return self._session.get(self._url, headers=headers, stream=True)

    def _merge(
        self, start: int, end: int, left: int, right: int
    ) -> typing.Generator[tuple[int, int]]:
        """Return a generator of intervals to be fetched.

        Args:
            start (int): Start of needed interval
            end (int): End of needed interval
            left (int): Index of first overlapping downloaded data
            right (int): Index after last overlapping downloaded data
        """
        lslice, rslice = self._left[left:right], self._right[left:right]
        i = start = min([start] + lslice[:1])
        end = max([end] + rslice[-1:])
        for j, k in zip(lslice, rslice):
            if j > i:
                yield i, j - 1
            i = k + 1
        if i <= end:
            yield i, end
        self._left[left:right], self._right[left:right] = [start], [end]

    def _download(self, start: int, end: int) -> None:
        """Download bytes from start to end inclusively."""
        with self._stay():
            left = bisect.bisect_left(self._right, start)
            right = bisect.bisect_right(self._left, end)
            for start, end in self._merge(start, end, left, right):
                response = self._stream_response(start, end)
                response.raise_for_status()
                self.seek(start)
                for chunk in response.iter_content(self._chunk_size):
                    self._file.write(chunk)


# https://stackoverflow.com/a/66733795
def compare_ast(
    node1: ast.expr | list[ast.expr], node2: ast.expr | list[ast.expr]
) -> bool:
    if type(node1) is not type(node2):
        return False

    if isinstance(node1, ast.AST):
        for k, v in vars(node1).items():
            if k in {"lineno", "end_lineno", "col_offset", "end_col_offset", "ctx"}:
                continue
            if not compare_ast(v, getattr(node2, k)):
                return False
        return True

    elif isinstance(node1, list) and isinstance(node2, list):
        return all(
            compare_ast(n1, n2) for n1, n2 in itertools.zip_longest(node1, node2)
        )
    else:
        return node1 == node2


def run():
    with requests.get(
        "https://pypi.org/simple/pyside6",
        headers={"Accept": "application/vnd.pypi.simple.v1+json"},
    ) as resp:
        resp.raise_for_status()

        data = resp.json()

    versions: list[str] = []
    for entry in data["files"]:
        if not entry["filename"].endswith(".whl"):
            continue

        name, version, buildtag, tags = packaging.utils.parse_wheel_filename(
            entry["filename"]
        )
        if version.pre:
            continue

        if not any(
            tag.platform.startswith("win_") and not tag.interpreter.startswith("pp")
            for tag in tags
        ):
            continue

        print(entry["filename"])

        # Store raw files in patches/data/<wheel>
        # This will allow us ot inspect them before deciding on how
        # to create patches.

        directory = os.path.join("patches", "data", str(version))
        os.makedirs(directory, exist_ok=True)

        session = requests.Session()
        wheel = LazyZipOverHTTP(entry["url"], session)
        with zipfile.ZipFile(wheel) as zf:
            for info in zf.infolist():
                if info.filename != "PySide6/__init__.py":
                    continue

                with open(
                    os.path.join(directory, os.path.basename(info.filename)), "wb"
                ) as f:
                    f.write(zf.read(info))
                break

        versions.append(str(version))

    print("Comparing files")
    first = versions.pop(0)

    while len(versions) > 1:
        leftFile = f"patches/data/{versions[0]}/__init__.py"
        rightFile = f"patches/data/{versions[1]}/__init__.py"
        with open(leftFile) as lfh, open(rightFile) as rfh:
            lhs = ast.parse(lfh.read())
            rhs = ast.parse(rfh.read())

        leftAST = next(
            node
            for node in lhs.body
            if isinstance(node, ast.FunctionDef)
            and node.name == "_additional_dll_directories"
        )

        rightAST = next(
            node
            for node in rhs.body
            if isinstance(node, ast.FunctionDef)
            and node.name == "_additional_dll_directories"
        )

        if not compare_ast(leftAST, rightAST):
            print(
                f"{versions[0]} and {versions[1]}'s _additional_dll_directories function differ"
            )
            leftCode = ast.unparse(leftAST).splitlines(keepends=True)
            rightCode = ast.unparse(rightAST).splitlines(keepends=True)

            result = difflib.unified_diff(
                leftCode, rightCode, fromfile=leftFile, tofile=rightFile
            )

            sys.stdout.writelines(result)

        versions.pop(0)


run()



# SPDX-FileCopyrightText: 2022 Contributors to the rez project
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import os
import sys
import json
import typing
import logging
import tempfile
import itertools
import subprocess
import dataclasses

import dataclasses_json

import rez_pip.data
import rez_pip.plugins
import rez_pip.exceptions

if typing.TYPE_CHECKING:
    import rez_pip.compat

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Metadata(dataclasses_json.DataClassJsonMixin):
    """Represents metadata for a package"""

    version: str
    name: str


@dataclasses.dataclass
class ArchiveInfo(dataclasses_json.DataClassJsonMixin):
    #: Archive hash
    hash: str

    #: Archive hashes
    hashes: typing.Dict[str, str]


@dataclasses.dataclass
class DownloadInfo(dataclasses_json.DataClassJsonMixin):
    #: Download URL
    url: str

    #: Archive information
    archive_info: ArchiveInfo

    dataclass_json_config = dataclasses_json.config(
        undefined=dataclasses_json.Undefined.EXCLUDE
    )


@dataclasses.dataclass(frozen=True)
class PackageInfo(dataclasses_json.DataClassJsonMixin):
    """Represents data returned by pip for a single package"""

    #: Download information
    download_info: DownloadInfo

    #: Is this a direct dependency?
    is_direct: bool

    #: Is this a requested package?
    requested: bool

    #: Metadata about the package
    metadata: Metadata

    dataclass_json_config = dataclasses_json.config(
        undefined=dataclasses_json.Undefined.EXCLUDE
    )

    @property
    def name(self) -> str:
        """Package name"""
        return self.metadata.name

    @property
    def version(self) -> str:
        """Package version"""
        return self.metadata.version

    def isDownloadRequired(self) -> bool:
        return not self.download_info.url.startswith("file://")


@dataclasses.dataclass(frozen=True)  # Nonsense, but we have to do this here too...
class DownloadedArtifact(PackageInfo):
    """
    This is a subclass of :class:`PackageInfo`. It's used to represent a local wheel.
    It is immutable so that we can clearly express immutability in plugins.
    """

    _localPath: str

    @property
    def path(self) -> str:
        """Path to the package on disk."""
        if not self.isDownloadRequired():
            # It's a local file, so we can return the URL (without file://)
            return self.download_info.url[7:]

        return self._localPath


T = typing.TypeVar("T", PackageInfo, DownloadedArtifact)


class PackageGroup(typing.Generic[T]):
    """A group of package. The order of packages and dists must be the same."""

    #: List of packages
    packages: tuple[T, ...]

    #: List of distributions
    dists: list[rez_pip.compat.importlib_metadata.Distribution]

    # Using a tuple to make it immutable
    def __init__(self, packages: tuple[T, ...]) -> None:
        self.packages = packages
        self.dists = []

    def __str__(self) -> str:
        return "PackageGroup({})".format(
            [f"{p.name}=={p.version}" for p in self.packages]
        )

    def __repr__(self) -> str:
        return "PackageGroup({})".format(
            [f"{p.name}=={p.version}" for p in self.packages]
        )

    def __bool__(self) -> bool:
        return bool(self.packages)

    def __eq__(self, value: typing.Any) -> bool:
        """Needed for tests"""
        if not isinstance(value, PackageGroup):
            return False

        return self.packages == value.packages and self.dists == value.dists

    @property
    def downloadUrls(self) -> list[str]:
        """List of download URLs"""
        return [p.download_info.url for p in self.packages]


def getBundledPip() -> str:
    return os.path.join(os.path.dirname(rez_pip.data.__file__), "pip.pyz")


def getPackages(
    packageNames: list[str],
    pip: str,
    pythonVersion: str,
    pythonExecutable: str,
    requirements: list[str],
    constraints: list[str],
    extraArgs: list[str],
) -> list[PackageInfo]:
    rez_pip.plugins.getHook().prePipResolve(
        packages=tuple(packageNames), requirements=tuple(requirements)
    )

    _fd, tmpFile = tempfile.mkstemp(prefix="pip-install-output", text=True)
    os.close(_fd)
    # We can't with "with" (context manager) because it will fail on Windows.
    # Windows doesn't allow two different processes to write if the file is
    # already opened.
    try:
        command = [
            # We need to use the real interpreter because pip can't resolve
            # markers correctly even if --python-version is provided.
            # See https://github.com/pypa/pip/issues/11664.
            pythonExecutable,
            pip,
            "install",
            "-q",
            *packageNames,
            *list(itertools.chain(*zip(["-r"] * len(requirements), requirements))),
            *list(itertools.chain(*zip(["-c"] * len(constraints), constraints))),
            "--disable-pip-version-check",
            "--dry-run",
            "--ignore-installed",
            f"--python-version={pythonVersion}" if pythonVersion else "",
            "--only-binary=:all:",
            "--target=/tmp/asd",
            "--disable-pip-version-check",
            "--report",  # This is the "magic". Pip will generate a JSON with all the resolved URLs.
            tmpFile,
            *extraArgs,
        ]

        _LOG.debug(f"Running {' '.join(command)!r}")
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

        pipOutput = []
        while True:
            stdout = typing.cast(typing.IO[str], process.stdout).readline()
            if process.poll() is not None:
                break
            if stdout:
                pipOutput.append(stdout.rstrip())
                sys.stdout.write(stdout)

        if process.poll() != 0:
            output = "\n".join(pipOutput)
            raise rez_pip.exceptions.PipError(
                f"[bold red]Failed to run pip command[/]: {' '.join(command)!r}\n\n"
                "[bold]Pip reported this[/]:\n\n"
                f"{output}",
            )
        reportContent = _readPipReport(reportPath=tmpFile)
    finally:
        os.remove(tmpFile)

    rawPackages = reportContent["install"]

    packages: list[PackageInfo] = []

    for rawPackage in rawPackages:
        packageInfo = PackageInfo.from_dict(rawPackage)
        packages.append(packageInfo)

    rez_pip.plugins.getHook().postPipResolve(packages=tuple(packages))

    return packages


def _readPipReport(reportPath: str) -> dict[str, typing.Any]:
    """
    Retrieve the json report generated by pip as json dict object.
    """
    with open(reportPath, encoding="utf-8") as reportFile:
        reportContent: dict[typing.Any, typing.Any] = json.load(reportFile)
    return reportContent
